"""
File containing methods to define analytic flows, from known distributions from numpy or other non-differentialble functions.

This has still many things to do to be a fully functioning version of the main flow and should be used with care.
"""

###############################################################################
# initial imports and set-up:

import numpy as np
import scipy.optimize
import numdifftools as numdiff
from getdist import MCSamples

# tensorflow imports:
import tensorflow as tf

# relative imports:
from . import synthetic_probability as sp
from ..utilities import stats_utilities as stutils


###############################################################################
# wrapper for tensorflow probability distributions:


class tf_prob_wrapper():
    """Thin wrapper around TensorFlow Probability distributions for flow usage."""
    def __init__(self, dist, prec=tf.float32):
        """
        Initialize the wrapper.

        :param dist: TensorFlow Probability distribution instance.
        :param prec: dtype used for casting inputs.
        """
        self.dist = dist
        self.label = dist.name
        num_params = dist.event_shape.as_list()[0]
        self.names = ['p'+str(i) for i in range(num_params)]
        self.prec = prec

    def log_pdf(self, coord):
        """
        Evaluate the log density.

        :param coord: coordinate tensor.
        :returns: log probability of ``coord``.
        """
        return self.dist.log_prob(tf.cast(coord, self.prec))

    def sim(self, N):
        """
        Draw samples from the wrapped distribution.

        :param N: number of samples.
        :returns: sampled points.
        """
        return self.dist.sample(N)
    
###############################################################################
# analytic flow class:


class analytic_flow():
    """Analytic flow backed by a user-provided distribution with sampling and log-density methods."""
    
    def __init__(self, dist, name_tag=None, param_names=None, param_labels=None, lims=None):
        """
        Analytic Flow model. This is build with numpy functions in mind and can be easily adapted to other cases.

        dist needs to implement sampling with a method called dist.sim
        and a pdf value, with a method dist.pdf
        """

        # check if the input distribution has the correct methods:
        keys = ['sim']
        for _k in keys:
            if not (hasattr(dist, _k) and callable(getattr(dist, _k))):
                raise ValueError('Input distribution does not have the '+_k+' method.')
        
        # check pdf method:
        if not (hasattr(dist, 'log_pdf') and callable(getattr(dist, 'log_pdf'))):
            self.has_log_pdf = False
            if not (hasattr(dist, 'pdf') and callable(getattr(dist, 'pdf'))):
                raise ValueError('Input distribution does not have the pdf method nor the log_pdf method.')
        else:
            self.has_log_pdf = True
        
        # copy in distribution:
        self._dist = dist
        
        # initialize tedious informations, from distribution or not:

        # name tag:
        if name_tag is not None:
            self.name_tag = name_tag
        else:
            self.name_tag = self._dist.label

        # param names:
        if param_names is not None:
            self.param_names = param_names
        else:
            self.param_names = self._dist.names

        # param labels:
        if param_labels is not None:
            self.param_labels = param_labels
        else:
            self.param_labels = self._dist.labels

        # param ranges:
        if lims is not None:
            self.parameter_ranges = lims
        else:
            self.parameter_ranges = self._dist.lims

        # initialize other things:
        self.num_params = len(self._dist.names)

        # initialize derivatives:
        self._jacobian_logP = numdiff.Jacobian(lambda x: np.log(self._dist.pdf(x)))
        self._hessian_logP = numdiff.Hessian(lambda x: np.log(self._dist.pdf(x)))

        #
        return None

    def cast(self, v):
        """Cast values to the flow precision dtype."""
        return tf.cast(v, dtype=sp.prec)
    
    def sample(self, num_samples):
        """
        Draw samples from the analytic distribution.

        :param num_samples: number of samples to generate.
        :returns: sampled points as a TensorFlow tensor.
        """
        return self.cast(self._dist.sim(num_samples))

    def MCSamples(self, size, logLikes=True, **kwargs):
        """
        Return MCSamples object from the syntetic probability.

        :param size: number of samples
        :param logLikes: logical, whether to include log-likelihoods or not.
        """
        samples = self.sample(size)
        if logLikes:
            loglikes = -self.log_probability(samples)
        else:
            loglikes = None
        mc_samples = MCSamples(
            samples=samples.numpy(),
            loglikes=loglikes.numpy(),
            names=self.param_names,
            labels=self.param_labels,
            ranges=self.parameter_ranges,
            name_tag=self.name_tag,
            **stutils.filter_kwargs(kwargs, MCSamples)
            )
        #
        return mc_samples
    
    def log_probability(self, coord):
        """
        Evaluate the log probability of coordinates under the analytic distribution.

        :param coord: coordinates to evaluate.
        :returns: log probability tensor.
        """
        # digest input:
        if tf.is_tensor(coord):
            _coord = coord.numpy()
        else:
            _coord = coord
        # return:
        if self.has_log_pdf:
            return self.cast(self._dist.log_pdf(_coord))
        else:
            return self.cast(np.log(self._dist.pdf(_coord)))
    
    def log_probability_jacobian(self, coord):
        """
        Compute the Jacobian of the log probability at given coordinates.

        :param coord: coordinates to evaluate.
        :returns: Jacobian of the log density.
        """
        # digest input:
        if tf.is_tensor(coord):
            _coord = coord.numpy()
        else:
            _coord = coord
        # cheaply vectorize and branch:
        if self.has_log_pdf:
            if len(_coord.shape) > 1:
                return self.cast(np.array([scipy.optimize.approx_fprime(_c, lambda x: self._dist.log_pdf(x)) for _c in _coord]))
            else:
                return self.cast(scipy.optimize.approx_fprime(_coord, lambda x: self._dist.log_pdf(x)))
        else:
            if len(_coord.shape) > 1:
                return self.cast(np.array([scipy.optimize.approx_fprime(_c, lambda x: np.log(self._dist.pdf(x))) for _c in _coord]))
            else:
                return self.cast(scipy.optimize.approx_fprime(_coord, lambda x: np.log(self._dist.pdf(x))))
        
    def log_probability_hessian(self, coord):
        """
        Compute the Hessian of the log probability at given coordinates.

        :param coord: coordinates to evaluate.
        :returns: Hessian of the log density.
        """
        # digest input:
        if tf.is_tensor(coord):
            _coord = coord.numpy()
        else:
            _coord = coord
        # cheaply vectorize:
        if len(_coord.shape) > 1:
            return self.cast(np.array([self._hessian_logP(_c)[0] for _c in _coord]))
        else:
            return self.cast(self._hessian_logP(_coord))

    def reset_tensorflow_caches(self):
        """Placeholder to mirror API; no caches to reset for analytic flow."""
        return None



"""Tests for Gaussian tension utilities."""

#########################################################################################################
# Imports

import unittest
from unittest.mock import patch

import numpy as np
from getdist import MCSamples, loadMCSamples
from getdist.gaussian_mixtures import GaussianND

import tensiometer.gaussian_tension as gt

#########################################################################################################
# Shared test fixtures


def setup_test(target):
    """Set up shared Gaussian test fixtures.

    :param target: test case instance to receive attributes.
    :returns: ``None``.
    """
    target.n1 = 2
    target.n2 = 3
    target.mean1 = 1.0 * np.ones(target.n1)
    target.mean2 = 2.0 * np.ones(target.n2)
    target.cov1 = 1.0 * np.diag(np.ones(target.n1))
    target.cov2 = 2.0 * np.diag(np.ones(target.n2))
    target.param_names1 = [f"p{i}" for i in range(target.n1)]
    target.param_names2 = [f"p{i}" for i in range(target.n2)]
    target.gaussian_1 = GaussianND(target.mean1, target.cov1, names=target.param_names1)
    target.gaussian_2 = GaussianND(target.mean2, target.cov2, names=target.param_names2)
    target.chain_1 = target.gaussian_1.MCSamples(1000)
    target.chain_2 = target.gaussian_2.MCSamples(1000)
    target.chain_12 = target.gaussian_2.MCSamples(1000)
    target.gaussian_prior = GaussianND(target.mean2, 10.0 * target.cov2, names=target.param_names2)
    target.prior_chain = target.gaussian_prior.MCSamples(1000)

#########################################################################################################
# Helper tests


class TestHelpers(unittest.TestCase):

    """Gaussian tension helper test suite."""
    def setUp(self):
        """Set up test fixtures."""
        setup_test(self)

    def test_helpers(self):
        """Test helper utilities for parameter checks."""
        self.assertEqual(
            self.chain_1.getParamNames().getRunningNames(),
            gt._check_param_names(self.chain_1, param_names=None),
        )
        self.assertEqual(gt._check_param_names(self.chain_1, param_names=["p1"]), ["p1"])
        gt._check_chain_type(self.chain_1)

    def test_validation_errors(self):
        """Test error handling for invalid helper inputs."""
        with self.assertRaises(ValueError):
            gt._check_param_names(self.chain_1, param_names=["test"])
        with self.assertRaises(TypeError):
            gt._check_chain_type(self.gaussian_1)
        other_chain = GaussianND(np.zeros(1), np.eye(1), names=["x"]).MCSamples(10)
        with self.assertRaises(ValueError):
            gt.Q_DM(self.chain_1, other_chain)

#########################################################################################################
# Utility tests


class TestUtilities(unittest.TestCase):

    """Gaussian tension utility test suite."""
    def setUp(self):
        """Set up test fixtures."""
        setup_test(self)

    def test_get_prior_covariance(self):
        """Test prior covariance estimation with ranges."""
        self.chain_1.setRanges({"p0": [0.0, 1.0], "p1": [0.0, 1.0]})
        gt.get_prior_covariance(self.chain_1)
        gt.get_prior_covariance(self.chain_2)

    def test_get_neff(self):
        """Test effective sample size calculations."""
        assert np.allclose(gt.get_Neff(self.chain_1), 2.0)
        gt.get_Neff(self.chain_1, prior_chain=self.prior_chain)
        assert np.allclose(gt.get_Neff(self.chain_1, param_names=["p1"]), 1.0)
        assert np.allclose(gt.get_Neff(self.chain_1, prior_factor=1.0), 2.0)

    def test_gaussian_approximation(self):
        """Test Gaussian approximation helper."""
        gt.gaussian_approximation(self.chain_1)
        gt.gaussian_approximation(self.chain_1, param_names=["p1"])
        self.chain_1.label = "chain_1"
        temp = gt.gaussian_approximation(self.chain_1)
        assert temp.label == "Gaussian " + self.chain_1.label
        self.chain_1.label = None
        self.chain_1.name_tag = "chain_1"
        temp = gt.gaussian_approximation(self.chain_1)
        assert temp.label == "Gaussian_" + self.chain_1.name_tag

    def test_q_metrics(self):
        """Test Q_MAP and Q_DMAP metrics."""
        q_map, dof = gt.Q_MAP(self.chain_1, num_data=10)
        self.assertTrue(np.isfinite(q_map))
        q_dmap, dofs = gt.Q_DMAP(self.chain_1, self.chain_2, self.chain_12, prior_chain=self.prior_chain)
        self.assertTrue(np.isfinite(q_dmap))


#########################################################################################################
# Slow integration tests


class TestGaussianTensionSlow(unittest.TestCase):
    """Gaussian tension slow test suite."""
    def setUp(self):
        """Set up test fixtures."""
        names = ["p0", "p1"]
        self.chain_1 = gt.GaussianND(np.array([0.0, 0.0]), np.eye(2), names=names).MCSamples(500)
        self.chain_2 = gt.GaussianND(np.array([0.2, -0.1]), 1.5 * np.eye(2), names=names).MCSamples(500)
        self.chain_12 = gt.GaussianND(np.array([0.1, -0.05]), 0.8 * np.eye(2), names=names).MCSamples(500)
        self.prior_chain = gt.GaussianND(np.array([0.0, 0.0]), 5.0 * np.eye(2), names=names).MCSamples(500)

    def test_localized_neff_and_cutoff(self):
        """Test localized effective sample size and cutoff estimation."""
        neff_localized = gt.get_Neff(self.chain_1, prior_chain=self.prior_chain,
                                     param_names=["p0", "p1"], localize=True, scale=5.0)
        self.assertGreaterEqual(neff_localized, 0)
        cutoff, eig, eigv, names = gt.Q_UDM_get_cutoff(self.chain_1, self.chain_2, self.chain_12,
                                                       prior_chain=self.prior_chain,
                                                       param_names=["p0", "p1"],
                                                       prior_factor=0.5)
        self.assertTrue(cutoff >= 1.0)
        self.assertEqual(len(names), eigv.shape[0])

    def test_cpca_pipeline_and_print(self):
        """Test CPCA pipeline and printing."""
        fisher_1 = np.linalg.inv(self.chain_1.cov(pars=["p0", "p1"]))
        fisher_12 = np.linalg.inv(self.chain_12.cov(pars=["p0", "p1"]))
        res = gt.linear_CPCA(fisher_1, fisher_12, ["p0", "p1"],
                             conditional_params=["p0"],
                             marginalized_parameters=[],
                             dimensional_reduce=False,
                             normparam="p0",
                             dimensional_threshold=0.05)
        self.assertIn("CPCA_eig", res)
        chain_res = gt.linear_CPCA_chains(self.chain_1, self.chain_12, ["p0", "p1"])
        text_verbose = gt.print_CPCA_results(chain_res, verbose=True, num_modes=1)
        text_brief = gt.print_CPCA_results(chain_res, verbose=False, num_modes=1)
        self.assertIn("CPCA for", text_verbose)
        self.assertIn("CPCA for", text_brief)

    def test_q_functions(self):
        """Test Q_UDM and component helpers."""
        cutoff = gt.Q_UDM_get_cutoff(self.chain_1, self.chain_2, self.chain_12,
                                     prior_chain=self.prior_chain)[0]
        q_udm, dofs_udm = gt.Q_UDM(self.chain_1, self.chain_12, lower_cutoff=1.0,
                                   upper_cutoff=cutoff + 1.0)
        self.assertGreaterEqual(dofs_udm, 0)
        eig, eigv, names = gt.Q_UDM_KL_components(self.chain_1, self.chain_12, param_names=["p0", "p1"])
        self.assertEqual(len(names), eigv.shape[0])

#########################################################################################################
# Additional test helpers


def make_simple_chain():
    """Make a simple 2D MCSamples chain.

    :returns: chain with two parameters.
    """
    samples = np.array([[0.0, 1.0], [1.0, 2.0]])
    names = ["p0", "p1"]
    labels = ["p0", "p1"]
    chain = MCSamples(samples=samples, names=names, labels=labels)
    return chain


#########################################################################################################
# Additional tests


class TestGaussianTensionAdditional(unittest.TestCase):
    """Gaussian tension additional test suite."""
    def test_check_param_names_error(self):
        """Test parameter name validation errors."""
        chain = make_simple_chain()
        with self.assertRaises(ValueError):
            gt._check_param_names(chain, ["missing"])

    def test_check_common_names_error(self):
        """Test shared-name validation errors."""
        with self.assertRaises(ValueError):
            gt._check_common_names(["a"], ["b"])

    def test_get_prior_covariance_with_missing_ranges(self):
        """Test prior covariance with missing parameter ranges."""
        chain = make_simple_chain()
        cov = gt.get_prior_covariance(chain, param_names=["p0", "p1"])
        self.assertEqual(cov.shape, (2, 2))
        self.assertTrue(np.all(np.isfinite(cov)))

    def test_linear_cpca_errors(self):
        """Test linear CPCA validation errors."""
        fisher = np.eye(2)
        with self.assertRaises(ValueError):
            gt.linear_CPCA(fisher_1=fisher, fisher_12=np.eye(3), param_names=["p0", "p1", "p2"])
        with self.assertRaises(ValueError):
            gt.linear_CPCA(fisher_1=fisher, fisher_12=fisher, param_names=["p0", "p1"], conditional_params=["bad"])
        with self.assertRaises(ValueError):
            gt.linear_CPCA(fisher_1=fisher, fisher_12=fisher, param_names=["p0", "p1"], marginalized_parameters=["bad"])

    def test_linear_cpca_marginalization_and_normparam(self):
        """Test linear CPCA marginalization and normalization settings."""
        fisher1 = np.array([[2.0, 0.1], [0.1, 1.5]])
        fisher12 = np.array([[1.5, 0.0], [0.0, 1.0]])
        results = gt.linear_CPCA(
            fisher_1=fisher1,
            fisher_12=fisher12,
            param_names=["p0", "p1"],
            marginalized_parameters=["p1"],
            normparam="p0",
            dimensional_reduce=True,
            dimensional_threshold=0.0,
        )
        self.assertIn("CPCA_eig", results)
        self.assertIn("CPCA_projector", results)
        self.assertEqual(results["CPCA_eig"].shape[0], results["CPCA_eigv"].shape[0])

    def test_localized_covariance_param_error(self):
        """Test localized covariance parameter validation."""
        chain = make_simple_chain()
        with self.assertRaises(ValueError):
            gt.get_localized_covariance(chain, chain, param_names=["p0", "p1"], localize_params=["bad"])

    def test_q_dm_cutoff_and_prior_mismatch(self):
        """Test Q_DM cutoff and prior mismatch validation."""
        chain1 = make_simple_chain()
        chain2 = make_simple_chain()
        chain1.setRanges({"p0": [-1.0, 1.0], "p1": [-1.0, 1.0]})
        chain2.setRanges({"p0": [0.0, 1.0], "p1": [0.0, 1.0]})
        with self.assertRaises(ValueError):
            gt.Q_DM(chain1, chain2, cutoff=-0.1)
        with self.assertRaises(ValueError):
            gt.Q_DM(chain1, chain2, cutoff=0.1)

    def test_cpca_print_branches(self):
        """Test CPCA print formatting branches."""
        cpca = {
            "param_names": ["a", "b"],
            "param_labels": ["A", "B"],
            "conditional_params": ["cond"],
            "marginalized_parameters": ["marg"],
            "Neff": 1.5,
            "Neff_spectrum": np.array([1.0, 0.5]),
            "KL_divergence": 0.2,
            "KL_spectrum": np.array([0.2, 0.1]),
            "CPCA_eig": np.array([2.0, 0.5]),
            "CPCA_eigv": np.array([[1.0, 0.0], [0.0, 1.0]]),
            "CPCA_projector": np.array([[1.0, 0.0], [0.0, 0.5]]),
            "CPCA_var_contributions": np.array([[0.6, 0.1], [0.4, 0.0]]),
            "CPCA_var_filter": np.array([[True, False], [True, False]]),
            "normparam": 0,
            "reference_point": np.array([0.1, -0.2]),
            "correlation_mode_parameter": np.array([[0.1, 0.2], [0.3, 0.4]]),
            "correlation_parameter_names": ["c1", "c2"],
        }
        text_verbose = gt.print_CPCA_results(cpca, verbose=True, num_modes=None)
        text_brief = gt.print_CPCA_results(cpca, verbose=False, num_modes=1)
        self.assertIn("CPCA for", text_verbose)
        self.assertIn("CPCA for", text_brief)

    def test_q_udm_component_errors(self):
        """Test Q_UDM component validation errors."""
        chain = make_simple_chain()
        with self.assertRaises(ValueError):
            gt.Q_UDM_fisher_components(chain, chain, which="bad")
        with self.assertRaises(ValueError):
            gt.Q_UDM_covariance_components(chain, chain, which="bad")

    def test_get_map_loglike_paths(self):
        """Test MAP loglike extraction paths."""
        class BestFit:
            """Best Fit test suite."""
            def __init__(self):
                """Init."""
                self.chiSquareds = [("data", type("C", (), {"chisq": 2.0})())]
                self.logLike = 0.0

            def getParamDict(self):
                """Get Param Dict."""
                return {}

        class Stats:
            """Stats test suite."""
            def __init__(self):
                """Init."""
                self.logLike_sample = -1.0

            def list(self):
                """List."""
                return ["chi2_data", "chi2_prior"]

            def parWithName(self, name):
                """Par With Name."""
                return type("P", (), {"bestfit_sample": 4.0})()

        class DummyChain:
            """Dummy Chain test suite."""
            def getBestFit(self, max_posterior=True):
                """Get Best Fit."""
                return BestFit()

        class DummyChainFallback:
            """Dummy Chain Fallback test suite."""
            def getBestFit(self, max_posterior=True):
                """Get Best Fit."""
                raise RuntimeError("no best fit")

            def getLikeStats(self):
                """Get Like Stats."""
                return Stats()

        val = gt.get_MAP_loglike(DummyChain(), feedback=False)
        self.assertTrue(np.isfinite(val))
        self.assertEqual(BestFit().getParamDict(), {})
        val2 = gt.get_MAP_loglike(DummyChainFallback(), feedback=False)
        self.assertTrue(np.isfinite(val2))

#########################################################################################################
# Targeted test helpers


def make_chain(mean_offset=0.0, name_tag="c"):
    """Make a small chain with two parameters.

    :param mean_offset: offset applied to sample means.
    :param name_tag: chain name tag to attach.
    :returns: chain with ranges set for ``p0`` and ``p1``.
    """
    samples = np.array([[0.1 + mean_offset, 0.2 + mean_offset],
                        [0.4 + mean_offset, -0.3 + mean_offset],
                        [-0.2 + mean_offset, 0.5 + mean_offset]])
    chain = MCSamples(samples=samples, names=["p0", "p1"], labels=["p0", "p1"], name_tag=name_tag)
    chain.setRanges({"p0": [-1.0, 1.0], "p1": [-2.0, 2.0]})
    return chain


#########################################################################################################
# Targeted tests


class TestGaussianTensionTargeted(unittest.TestCase):
    """Gaussian tension targeted test suite."""
    def test_localized_covariance_edge_branches(self):
        """Test localized covariance edge branches."""
        c2 = make_chain(0.05)
        c2.setRanges({"p0": [-2.0, 2.0], "p1": [-2.0, 2.0]})

        zero_weight_chain = make_chain()
        zero_weight_chain.weights = np.zeros_like(zero_weight_chain.weights)
        cov = gt.get_localized_covariance(zero_weight_chain, c2, param_names=["p0", "p1"])
        self.assertEqual(cov.shape, (2, 2))

        single = make_chain()
        single.samples = single.samples[:1, :]
        single.weights = single.weights[:1]
        cov_single = gt.get_localized_covariance(single, c2, param_names=["p0", "p1"])
        self.assertEqual(cov_single.shape, (2, 2))

        tiny_weight_chain = make_chain()
        tiny_weight_chain.weights = np.full_like(tiny_weight_chain.weights, 1.0e-300)
        cov_mock = gt.get_localized_covariance(tiny_weight_chain, c2, param_names=["p0", "p1"])
        self.assertEqual(cov_mock.shape, (2, 2))

        with patch("tensiometer.gaussian_tension.np.cov", return_value=np.full((2, 2), np.nan)):
            cov_nan = gt.get_localized_covariance(make_chain(), c2, param_names=["p0", "p1"], scale=10.0)
            self.assertTrue(np.all(np.isfinite(cov_nan)))

    def test_localized_covariance_main_path(self):
        """Test localized covariance main path."""
        c1 = make_chain()
        c2 = make_chain(0.05)
        cov = gt.get_localized_covariance(c1, c2, param_names=["p0", "p1"], scale=2.0)
        self.assertEqual(cov.shape, (2, 2))
        mean = np.mean(c2.samples, axis=0)
        offsets = np.array([
            [0.0, 0.0],
            [0.03, -0.01],
            [-0.02, 0.04],
            [0.05, 0.02],
        ])
        near_samples = mean + offsets
        base = np.linspace(0, 5, 50)
        spread = np.column_stack([base, np.sin(base) + 0.1 * base])
        many_samples = np.vstack([spread, near_samples])
        heavy_chain = MCSamples(samples=many_samples, names=["p0", "p1"], labels=["p0", "p1"])
        heavy_chain.setRanges({"p0": [-10, 10], "p1": [-10, 10]})
        with patch("builtins.print") as mock_print:
            gt.get_localized_covariance(heavy_chain, c2, param_names=["p0", "p1"], scale=0.05)
            self.assertTrue(mock_print.called)
        cov2 = gt.get_localized_covariance(c1, c2, param_names=["p0", "p1"], localize_params=["p0"])
        self.assertEqual(cov2.shape[0], 2)

    def test_q_udm_covariance_components_invalid(self):
        """Test Q_UDM covariance components invalid branch."""
        c1 = make_chain()
        c2 = make_chain(0.05)
        with self.assertRaises(ValueError):
            gt.Q_UDM_covariance_components(c1, c2, which="bad")

    def test_q_dm_with_prior_chain(self):
        """Test Q_DM with and without a prior chain."""
        c1 = make_chain()
        c2 = make_chain(0.02)
        prior = make_chain(0.1)
        prior.setRanges({"p0": [-1, 1], "p1": [-2, 2]})
        q, dofs = gt.Q_DM(c1, c2, prior_chain=prior, cutoff=0.01, prior_factor=0.5)
        self.assertTrue(np.isfinite(q))
        self.assertGreaterEqual(dofs, 0)
        q2, dofs2 = gt.Q_DM(c1, c2, prior_chain=None, cutoff=0.01)
        self.assertTrue(np.isfinite(q2))
        self.assertGreaterEqual(dofs2, 0)
        with patch("tensiometer.gaussian_tension.stutils.KL_decomposition",
                   side_effect=[(np.array([2.0, 0.5]), np.eye(2)), (np.array([1.01, 1.02]), np.eye(2))]):
            q3, dofs3 = gt.Q_DM(c1, c2, prior_chain=prior, cutoff=0.01)
            self.assertTrue(np.isfinite(q3))
            self.assertGreaterEqual(dofs3, 0)

    def test_print_cpca_correlation_block(self):
        """Test CPCA correlation block formatting."""
        cpca = {
            "param_names": ["p0", "p1"],
            "param_labels": ["P0", "P1"],
            "conditional_params": ["p0"],
            "marginalized_parameters": [],
            "Neff": 1.0,
            "Neff_spectrum": np.array([1.0, 0.5]),
            "KL_divergence": 0.1,
            "KL_spectrum": np.array([0.1, 0.05]),
            "CPCA_eig": np.array([2.0, 1.1]),
            "CPCA_eigv": np.eye(2),
            "CPCA_projector": np.array([[1.0, 0.2], [0.1, 1.0]]),
            "CPCA_var_contributions": np.ones((2, 2)),
            "CPCA_var_filter": np.ones((2, 2), dtype=bool),
            "normparam": 0,
            "reference_point": np.array([0.0, 0.1]),
            "correlation_mode_parameter": np.array([[0.3, 0.4], [0.5, 0.6]]),
            "correlation_parameter_names": ["p0", "p1"],
        }
        text = gt.print_CPCA_results(cpca, verbose=True, num_modes=2)
        self.assertIn("correlations", text)
        cpca.pop("param_labels")
        cpca["CPCA_var_filter"] = np.array([[True, False], [False, True]])
        text2 = gt.print_CPCA_results(cpca, verbose=False, num_modes=1)
        self.assertIn("CPCA for", text2)
        cpca["CPCA_eig"] = np.array([0.9, 1.0])
        cpca["CPCA_var_contributions"] = np.zeros((2, 2))
        text3 = gt.print_CPCA_results(cpca, verbose=True, num_modes=2)
        self.assertIn("noisy", text3)
        cpca["conditional_params"] = []
        cpca["marginalized_parameters"] = []
        text4 = gt.print_CPCA_results(cpca, verbose=True, num_modes=1)
        self.assertIn("CPCA for", text4)
        log_flags = []

        class Toggle(float):
            """Toggle test suite."""
            counter = 0
            def __new__(cls, value):
                """New."""
                obj = float.__new__(cls, value)
                return obj
            def __sub__(self, other):
                """Sub."""
                return self
            def __gt__(self, other):
                """Gt."""
                Toggle.counter += 1
                res = False if Toggle.counter == 6 else True
                log_flags.append(res)
                return res

        cpca_small = {
            "param_names": ["p0"],
            "param_labels": ["A"],
            "conditional_params": [],
            "marginalized_parameters": [],
            "Neff": 1.0,
            "Neff_spectrum": np.array([1.0]),
            "KL_divergence": 0.1,
            "KL_spectrum": np.array([0.1]),
            "CPCA_eig": np.array([Toggle(2.0)], dtype=object),
            "CPCA_eigv": np.array([[1.0]]),
            "CPCA_projector": np.array([[1.0]]),
            "CPCA_var_contributions": np.array([[1.0]]),
            "CPCA_var_filter": np.array([[True]]),
            "normparam": None,
        }
        text5 = gt.print_CPCA_results(cpca_small, verbose=True, num_modes=1)
        self.assertIn("parameter combinations", text5)
        self.assertIn(False, log_flags)
        self.assertTrue(log_flags[-1] is False)

    def test_q_udm_get_cutoff_success(self):
        """Test Q_UDM cutoff bracketing success and failure."""
        dummy_names = ["p0", "p1"]
        with patch("tensiometer.gaussian_tension._check_param_names", return_value=dummy_names), \
                patch("tensiometer.gaussian_tension.Q_UDM_KL_components",
                      return_value=(np.array([2.0, 3.0]), np.eye(2), dummy_names)), \
                patch("tensiometer.gaussian_tension.get_Neff", side_effect=[1.0, 1.0, 1.0]):
            cutoff, eig, eigv, names = gt.Q_UDM_get_cutoff(make_chain(), make_chain(), make_chain())
        self.assertGreater(cutoff, 0)
        self.assertEqual(list(names), dummy_names)
        # failure branch when cutoff cannot be bracketed
        with patch("tensiometer.gaussian_tension.Q_UDM_KL_components",
                   return_value=(np.array([1.1, 1.1]), np.eye(2), dummy_names)), \
                patch("tensiometer.gaussian_tension.get_Neff", side_effect=[5.0, 5.0, 5.0]):
            with self.assertRaises(ValueError):
                gt.Q_UDM_get_cutoff(make_chain(), make_chain(), make_chain())
        c1 = MCSamples(samples=np.array([[0.1]]), names=["a"], labels=["a"])
        c2 = MCSamples(samples=np.array([[0.2]]), names=["b"], labels=["b"])
        with self.assertRaises(ValueError):
            gt.Q_UDM_get_cutoff(c1, c2, c2)

    def test_q_udm_fisher_covariance_branches(self):
        """Test Q_UDM fisher and covariance branches."""
        dummy_names = ["p0", "p1"]
        eig = np.array([2.0, 3.0])
        eigv = np.array([[1.0, 0.0], [0.0, 1.0]])
        with patch("tensiometer.gaussian_tension.Q_UDM_KL_components", return_value=(eig, eigv, dummy_names)):
            names, eig_out, frac_fisher, fisher = gt.Q_UDM_fisher_components(make_chain(), make_chain(), which="2")
            self.assertEqual(names, dummy_names)
            self.assertEqual(eig_out.tolist(), eig.tolist())
            names, eig_out, frac_fisher, fisher = gt.Q_UDM_fisher_components(make_chain(), make_chain(), which="12")
            self.assertEqual(frac_fisher.shape[0], 2)
            names, eig_out, frac_cov = gt.Q_UDM_covariance_components(make_chain(), make_chain(), which="12")
            self.assertEqual(frac_cov.shape[0], 2)
            names, eig_out, frac_fisher, fisher = gt.Q_UDM_fisher_components(make_chain(), make_chain(), which="1")
            self.assertTrue(np.all(np.isfinite(frac_fisher)))
            names, eig_out, frac_cov = gt.Q_UDM_covariance_components(make_chain(), make_chain(), which="2")
            self.assertTrue(np.all(np.isfinite(frac_cov)))
            names, eig_out, frac_cov = gt.Q_UDM_covariance_components(make_chain(), make_chain(), which="1")
            self.assertTrue(np.all(np.isfinite(frac_cov)))

    def test_q_udm_components_errors(self):
        """Test Q_UDM component errors for missing inputs."""
        c1 = MCSamples(samples=np.array([[0.1]]), names=["a"], labels=["a"])
        c2 = MCSamples(samples=np.array([[0.2]]), names=["b"], labels=["b"])
        with self.assertRaises(ValueError):
            gt.Q_UDM_KL_components(c1, c2)
        with self.assertRaises(ValueError):
            gt.Q_UDM_fisher_components(c1, c2)
        with self.assertRaises(ValueError):
            gt.Q_UDM_covariance_components(c1, c2)
        with patch("tensiometer.gaussian_tension.Q_UDM_KL_components",
                   return_value=(np.array([1.5, 0.5]), np.eye(2), ["p0", "p1"])):
            q, dofs = gt.Q_UDM(make_chain(), make_chain())
            self.assertTrue(np.isfinite(q))
            self.assertGreaterEqual(dofs, 1)

    def test_get_map_loglike_bestfit_empty_chi(self):
        """Exercise MAP loglike fallbacks when best-fit data are missing."""

        class EmptyChiBestFit:
            """Best-fit mock without chi-squared entries."""

            def __init__(self):
                """Initialize best-fit values."""
                self.chiSquareds = []
                self.logLike = 5.0

            def getParamDict(self):
                """Return prior contribution."""
                return {"prior": 1.0}

        class ChainWithPriorBestFit:
            """Chain exposing explicit best-fit data with a prior term."""

            def getBestFit(self, max_posterior=True):
                """Return stored best-fit state."""
                return EmptyChiBestFit()

        self.assertEqual(gt.get_MAP_loglike(ChainWithPriorBestFit(), feedback=False), -2.0)

        class EmptyStats:
            """Stats container with no chi-squared entries."""

            def list(self):
                """Return an empty chi-squared list."""
                return []

            @property
            def logLike_sample(self):
                """Provide the sampled log-likelihood."""
                return -2.5

        class ChainWithStatsFallback:
            """Chain that triggers the stats fallback path."""

            def getBestFit(self, max_posterior=True):
                """Raise to force the fallback path."""
                raise RuntimeError("fail")

            def getLikeStats(self):
                """Return stats with no chi-squared labels."""
                return EmptyStats()

        self.assertEqual(gt.get_MAP_loglike(ChainWithStatsFallback(), feedback=False), 1.25)

        class ChiObj:
            """Container for a chi-squared value."""

            def __init__(self, value):
                """Store the chi-squared value."""
                self.chisq = value

        class ChiBestFit:
            """Best-fit mock with chi-squared entries."""

            def __init__(self):
                """Create chi-squared entries."""
                self.chiSquareds = [("d", ChiObj(4.0)), ("d2", ChiObj(2.0))]
                self.logLike = 0.0

            def getParamDict(self):
                """Return an empty parameter mapping."""
                return {}

        class ChainWithChiBestFit:
            """Chain returning best-fit info with chi-squared values."""

            def getBestFit(self, max_posterior=True):
                """Return chi-squared best-fit entries."""
                return ChiBestFit()

        self.assertEqual(gt.get_MAP_loglike(ChainWithChiBestFit(), feedback=False), -3.0)
        self.assertEqual(ChiBestFit().getParamDict(), {})

        class StatsWithPrior:
            """Stats including a prior chi-squared term."""

            def list(self):
                """Return chi-squared labels with a prior entry."""
                return ["chi2_data", "chi2_prior"]

            def parWithName(self, name):
                """Return parameter info for a chi-squared entry."""

                class Param:
                    """Parameter wrapper holding a best-fit sample."""

                    def __init__(self, value):
                        """Set the best-fit sample value."""
                        self.bestfit_sample = value

                return Param(1.0)

        class ChainWithPriorStats:
            """Chain raising on best fit and falling back to stats with prior."""

            def getBestFit(self, max_posterior=True):
                """Raise to exercise fallback."""
                raise RuntimeError("fail")

            def getLikeStats(self):
                """Return stats with prior chi-squared entries."""
                return StatsWithPrior()

        self.assertEqual(gt.get_MAP_loglike(ChainWithPriorStats(), feedback=False), -0.5)

        class NoStatsChain:
            """Chain raising on best fit and returning no stats."""

            def getBestFit(self, max_posterior=True):
                """Raise to trigger missing stats fallback."""
                raise RuntimeError("fail")

            def getLikeStats(self):
                """Return no stats information."""
                return None

        self.assertEqual(gt.get_MAP_loglike(NoStatsChain(), feedback=False), 0.0)

        class StatsWithoutPrior:
            """Stats object with multiple chi-squared entries and no prior."""

            def list(self):
                """Return chi-squared labels without a prior entry."""
                return ["chi2_data", "chi2_extra"]

            def parWithName(self, name):
                """Return parameter info for a chi-squared entry."""

                class Param:
                    """Parameter wrapper holding a best-fit sample."""

                    def __init__(self, value):
                        """Set the best-fit sample value."""
                        self.bestfit_sample = value

                return Param(2.0)

        class ChainWithStatsOnly:
            """Chain raising on best fit and using stats without prior."""

            def getBestFit(self, max_posterior=True):
                """Raise to force stats-only path."""
                raise RuntimeError("fail")

            def getLikeStats(self):
                """Return stats without prior information."""
                return StatsWithoutPrior()

        self.assertEqual(gt.get_MAP_loglike(ChainWithStatsOnly(), feedback=False), -2.0)

        class ChainPrintingWarning:
            """Chain exercising feedback path without stats."""

            def getBestFit(self, max_posterior=True):
                """Raise to emit warning when feedback is enabled."""
                raise RuntimeError("fail")

            def getLikeStats(self):
                """Return no stats information."""
                return None

        self.assertEqual(gt.get_MAP_loglike(ChainPrintingWarning(), feedback=True), 0.0)

        class BestFitWithoutPrior:
            """Best-fit mock without chi-squared entries or priors."""

            def __init__(self):
                """Initialize best-fit values."""
                self.chiSquareds = []
                self.logLike = 3.0

            def getParamDict(self):
                """Return an empty parameter dictionary."""
                return {}

        class ChainWithoutPrior:
            """Chain returning a best fit with no prior contributions."""

            def getBestFit(self, max_posterior=True):
                """Return best-fit values without priors."""
                return BestFitWithoutPrior()

        self.assertEqual(gt.get_MAP_loglike(ChainWithoutPrior(), feedback=False), -1.5)

    def test_get_neff_localize_and_linear_cpca_chains(self):
        """Test Neff localization and linear CPCA chain helper."""
        chain = make_chain()
        prior = make_chain()
        prior.setRanges({"p0": [-1, 1], "p1": [-2, 2]})
        neff = gt.get_Neff(chain, prior_chain=prior, localize=True, param_names=["p0", "p1"])
        self.assertTrue(np.isfinite(neff))
        res = gt.linear_CPCA_chains(chain, chain, param_names=["p0", "p1"], dimensional_reduce=False)
        self.assertIn("CPCA_eig", res)
        fisher = np.eye(2)
        res_lin = gt.linear_CPCA(fisher, fisher, ["p0", "p1"], dimensional_reduce=False)
        self.assertIn("CPCA_projector", res_lin)

    def test_linear_cpca_validation_errors(self):
        """Test linear CPCA validation errors."""
        fisher = np.eye(2)
        with self.assertRaises(ValueError):
            gt.linear_CPCA(fisher, np.eye(3), ["p0", "p1"])
        with self.assertRaises(ValueError):
            gt.linear_CPCA(fisher, fisher, ["p0", "p1"], conditional_params=["bad"])
        with self.assertRaises(ValueError):
            gt.linear_CPCA(fisher, fisher, ["p0", "p1"], marginalized_parameters=["bad"])
        res = gt.linear_CPCA(fisher, fisher, ["p0", "p1"],
                             conditional_params=["p1"], marginalized_parameters=[])
        self.assertIn("CPCA_eig", res)

    def test_print_cpca_verbose_blocks(self):
        """Test CPCA verbose print blocks."""
        cpca = {
            "param_names": ["p0", "p1"],
            "param_labels": ["A", "B"],
            "conditional_params": ["p1"],
            "marginalized_parameters": ["p0"],
            "Neff": 1.0,
            "Neff_spectrum": np.array([0.5, 0.5]),
            "KL_divergence": 0.1,
            "KL_spectrum": np.array([0.05, 0.05]),
            "CPCA_eig": np.array([2.0, 0.8]),
            "CPCA_eigv": np.eye(2),
            "CPCA_projector": np.eye(2),
            "CPCA_var_contributions": np.ones((2, 2)),
            "CPCA_var_filter": np.ones((2, 2), dtype=bool),
            "normparam": None,
        }
        text = gt.print_CPCA_results(cpca, verbose=True, num_modes=2)
        self.assertIn("fixed parameters", text)

#########################################################################################################
# Script entry point


if __name__ == "__main__":  # pragma: no cover
    unittest.main(verbosity=2)



"""
This file contains some utilities that are useful to run a function inside an independent sub-process.

When the subprocess is finished the memory is fully released, which is useful to avoid memory leaks, especially from tensorflow.
"""

###########################################################################################
# Initial imports:

import copy
import datetime
import multiprocessing as mp
import time
from functools import wraps

import psutil

###########################################################################################
# Default Settings:

# These settings define the default behavior of the `run_in_process` decorator.

default_settings = {
    'subprocess': True,         # Whether to run the function in a subprocess.
    'feedback_level': 1,        # Level of feedback provided:
                                # 0 - None, 1 - Minimal, 2 - Medium, 3 - Full.
    'context': 'fork',          # Multiprocessing context: 'fork', 'spawn', or 'forkserver'.
    'monitoring': True,         # Enable or disable monitoring of the subprocess.
    'monitoring_frequency': 1,  # Frequency of monitoring updates, in seconds.
    'timeout': None,            # Maximum allowed runtime in seconds. None means no timeout.
                                # The timeout is rounded to the nearest monitoring frequency.
}

###########################################################################################
# Hard-Coded Settings:

# These constants define formatting options for feedback messages.

feedback_offset = '  '         # Indentation for feedback messages.
feedback_offset_2 = '   | '    # Indentation for nested feedback messages.
feedback_separator = '****************************************************************'
                                # Separator line for formatting feedback output.

###########################################################################################
# `run_in_process` Decorator:
# This decorator modifies a function to optionally run it in a separate subprocess
# with configurable feedback and monitoring.

def run_in_process(**kwargs):
    """
    Decorator to run a function in a subprocess with optional feedback,
    monitoring, and timeout capabilities.

    :param subprocess: whether to execute in a subprocess.
    :param feedback_level: verbosity level from 0 to 3.
    :param context: multiprocessing context (``fork``, ``spawn`` or ``forkserver``).
    :param monitoring: enable monitoring of the subprocess.
    :param monitoring_frequency: interval, in seconds, for monitoring updates.
    :param timeout: maximum runtime in seconds; enables monitoring when set.
    :returns: wrapped function with subprocess capabilities.
    :raises ValueError: if supplied configuration values are invalid.
    """
    # Update settings with the provided overrides.
    settings = copy.deepcopy(default_settings)
    for _k in kwargs.keys():
        if _k in settings.keys():
            settings[_k] = kwargs[_k]

    # Validate settings.
    if not isinstance(settings['subprocess'], bool):
        raise ValueError('subprocess must be a boolean.')
    if not isinstance(settings['feedback_level'], int):
        raise ValueError('feedback_level must be an integer.')
    if not settings['feedback_level'] in [0, 1, 2, 3]:
        raise ValueError('feedback_level must be 0, 1, 2, or 3.')
    if not settings['context'] in ['fork', 'spawn', 'forkserver']:
        raise ValueError("context must be 'fork', 'spawn', or 'forkserver'.")
    if not isinstance(settings['monitoring'], bool):
        raise ValueError('monitoring must be a boolean.')
    if not isinstance(settings['monitoring_frequency'], int):
        raise ValueError('monitoring_frequency must be an integer (in seconds).')
    if settings['timeout'] is not None and not isinstance(settings['timeout'], int):
        raise ValueError('timeout must be an integer (in seconds).')

    # Expand settings into individual variables.
    subprocess = settings['subprocess']
    feedback_level = settings['feedback_level']
    context = settings['context']
    monitoring = settings['monitoring']
    monitoring_frequency = settings['monitoring_frequency']
    timeout = settings['timeout']

    # If a timeout is set, monitoring must be enabled.
    if timeout is not None:
        monitoring = True

    def decorator(func):
        """
        Inner decorator to wrap the target function.

        :param func: target function to decorate.
        :returns: wrapped function respecting the configured subprocess options.
        """
        if not subprocess:
            # If not running in a subprocess, return the original function.
            return func

        @wraps(func)
        def wrapper(*args, **kwargs):
            """
            Wrapper function to run the target function in a subprocess.

            :param args: positional arguments for the target function.
            :param kwargs: keyword arguments for the target function.
            :returns: result of the target function.
            """
            # Record the start time of the process.
            global_start_time = datetime.datetime.now()
            func_name = func.__name__
            result = None

            # Provide initial feedback based on the feedback level.
            if feedback_level > 0:
                print(feedback_separator, flush=True)
                print(f'* Running subprocess for function: {func_name}', flush=True)
                print(feedback_offset + f'Start time: {global_start_time.strftime("%Y-%m-%d %H:%M:%S")}', flush=True)
                print(feedback_separator, flush=True)
                if feedback_level > 1:
                    print('* Settings:', flush=True)
                    for key, value in settings.items():
                        print(feedback_offset + f'- {key}: {value}', flush=True)
                    print(feedback_separator, flush=True)
                if feedback_level > 2:
                    print('* Function arguments      :', args, flush=True)
                    print('* Function keyword args   :', kwargs, flush=True)
                    print(feedback_separator, flush=True)

            # Define the target function to be executed in the subprocess.
            def target_function(pipe_conn, *args, **kwargs):
                """
                Execute the target function in a subprocess and send the result back.

                :param pipe_conn: pipe connection for inter-process communication.
                :param args: positional arguments for the target function.
                :param kwargs: keyword arguments for the target function.
                """
                try:
                    # Execute the function and send the result through the pipe.
                    result = func(*args, **kwargs)
                    pipe_conn.send(result)
                except Exception as e:
                    # Send exceptions through the pipe for error handling.
                    import sys
                    import traceback
                    exc_type, exc_value, exc_traceback = sys.exc_info()
                    traceback_string = ''.join(traceback.format_exception(exc_type, exc_value, exc_traceback))
                    e.traceback = traceback_string
                    pipe_conn.send(e)
                finally:
                    pipe_conn.close()

            # Set up and start the subprocess.
            ctx = mp.get_context(context)
            parent_conn, child_conn = ctx.Pipe()
            
            # create the process:
            process = ctx.Process(target=target_function, args=(child_conn, *args), kwargs=kwargs)
            if feedback_level > 2:
                print('* Sub-process created.', flush=True)
            
            process.start()
            if feedback_level > 2:
                print('* Sub-process started.', flush=True)    

            # initial memory usage:
            if monitoring:
                initial_memory = psutil.Process(process.pid).memory_info().rss / 1024 / 1024
                if feedback_level > 1:
                    print('* Initial memory usage:', initial_memory, 'MB', flush=True)
                    print(feedback_separator, flush=True)
                peak_memory = initial_memory / 1024 / 1024
                    
            # Monitor the process and handle timeout.
            initial_time = time.time() if monitoring else None
 
            while True:

                # exit by process status:
                _process_status = process.is_alive()
                if feedback_level > 2:
                    print('* Process is alive:', _process_status, flush=True)
                if not _process_status:
                    if feedback_level > 2:
                        print(f'* Process exited with code {process.exitcode}.', flush=True)
                    if process.exitcode == 0:
                        break
                    else:
                        raise Exception('Process exited with code %d' % process.exitcode)                    

                # monitoring:
                if monitoring:
                    current_memory = psutil.Process(process.pid).memory_info().rss / 1024 / 1024
                    if feedback_level > 2:
                        print('* Current memory usage:', current_memory, 'MB', flush=True)
                    peak_memory = max(peak_memory, current_memory)

                # break if pipe is full:
                if parent_conn.poll():
                    if feedback_level > 2:
                        print('* Breaking, sub-process pipe is full.', flush=True)
                    break

                # break by timeout:
                if timeout and time.time() - initial_time > timeout:
                    process.terminate()
                    raise TimeoutError("Process timed out.")

                # sleep for monitoring frequency:
                time.sleep(monitoring_frequency)

            # Receive the result from the pipe
            result = parent_conn.recv()
            if feedback_level > 2:
                print('* Result received.', flush=True)

            # Wait for the process to finish and get the result:
            process.join()
            if feedback_level > 2:
                print('* Process joined.', flush=True)

            # monitoring stats if needed:
            if monitoring:
                final_time = time.time()
                total_time = final_time - initial_time
                if feedback_level > 0:
                    print(feedback_separator, flush=True)
                    print(f'* Total time elapsed: {total_time:.2f} seconds', flush=True)
                    print(f'* Peak memory usage: {peak_memory:.2f} MB', flush=True)
                    print(feedback_separator, flush=True)

            # Raise exceptions received from the subprocess.
            if isinstance(result, Exception):
                raise result

            return result

        return wrapper

    return decorator



"""Tests for parameter-difference tension calculations."""

#########################################################################################################
# Imports

import copy
import os
import unittest

import numpy as np
from getdist import MCSamples, WeightedSamples, loadMCSamples

import tensiometer.mcmc_tension.param_diff as pd

#########################################################################################################
# Test cases


class TestMcmcParamDiff(unittest.TestCase):

    """MCMC parameter-difference test suite."""
    def setUp(self):
        """Set up test fixtures."""
        self.here = os.path.dirname(os.path.abspath(__file__))
        self.chain_1 = loadMCSamples(self.here + "/../../test_chains/DES")
        self.chain_2 = loadMCSamples(self.here + "/../../test_chains/Planck18TTTEEE")
        self.chain_12 = loadMCSamples(self.here + "/../../test_chains/Planck18TTTEEE_DES")
        self.chain_prior = loadMCSamples(self.here + "/../../test_chains/prior")
        self.chain_1.getConvergeTests()
        self.chain_2.getConvergeTests()
        self.chain_12.getConvergeTests()
        self.chain_prior.getConvergeTests()
        self.chain_1.weighted_thin(int(self.chain_1.indep_thin))
        self.chain_2.weighted_thin(int(self.chain_2.indep_thin))
        self.chain_12.weighted_thin(int(self.chain_12.indep_thin))
        self.chain_prior.weighted_thin(int(self.chain_prior.indep_thin))

    def test_param_diff_with_fixed_param(self):
        """Test parameter differences with a fixed parameter."""
        names = [p.name for p in self.chain_1.getParamNames().names if p.name != "tau"]
        indices = [self.chain_1.index[n] for n in names]
        labels = [self.chain_1.getParamNames().parWithName(n).label for n in names]
        samples = self.chain_1.samples[:, indices]
        self.chain_1_mod = MCSamples(samples=samples,
                                     names=names,
                                     labels=labels,
                                     weights=self.chain_1.weights,
                                     loglikes=self.chain_1.loglikes,
                                     ranges=self.chain_1.ranges,
                                     ignore_rows=self.chain_1.ignore_rows,
                                     sampler=self.chain_1.sampler)
        self.chain_1_mod.updateBaseStatistics()
        self.diff_chain = pd.parameter_diff_chain(self.chain_1_mod,
                                                  self.chain_2,
                                                  boost=1,
                                                  fixed_params={"tau": 0.05})

#########################################################################################################
# Helper utilities


def make_samples(names=("p0", "p1"), weights=None, values=None, sampler="uncorrelated", name_tag=None, label=None):
    """Build a small MCSamples instance for parameter-difference tests.

    :param names: parameter names.
    :param weights: sample weights.
    :param values: sample values to use.
    :param sampler: sampler name to set.
    :param name_tag: chain name tag.
    :param label: chain label.
    :returns: configured MCSamples instance.
    """
    weights = np.ones(3) if weights is None else weights
    if values is None:
        values = np.arange(len(weights) * len(names)).reshape(len(weights), len(names)) * 0.1
    labels = [n.upper() for n in names]
    return MCSamples(samples=np.array(values), weights=np.array(weights), names=list(names),
                     labels=labels, sampler=sampler, ignore_rows=0, name_tag=name_tag, label=label)

#########################################################################################################
# Additional tests


class TestParamDiffAdditional(unittest.TestCase):
    """Parameter difference additional test suite."""
    def test_parameter_diff_weighted_samples_branches(self):
        """Test weighted-sample parameter difference branches."""
        ws1 = WeightedSamples(samples=np.array([[0.0, 0.5], [1.0, 1.5]]),
                              weights=np.array([1.0, 2.0]),
                              loglikes=np.array([0.1, 0.2]))
        ws1.name_tag = "c1"
        ws1.label = "C1"
        ws1.min_weight_ratio = 0.4
        ws2 = WeightedSamples(samples=np.array([[0.2, -0.5], [1.2, -1.0], [2.2, -1.5]]),
                              weights=np.array([1.0, 1.0, 1.0]))
        ws2.name_tag = "c2"
        ws2.label = "C2"
        ws2.min_weight_ratio = 0.3
        with self.assertRaises(TypeError):
            pd.parameter_diff_weighted_samples("bad", ws2)
        with self.assertRaises(TypeError):
            pd.parameter_diff_weighted_samples(ws1, "bad")
        with self.assertRaises(ValueError):
            pd.parameter_diff_weighted_samples(ws1, ws2, indexes_1=[0], indexes_2=[0, 1])
        res = pd.parameter_diff_weighted_samples(ws1, ws2, boost=None)
        self.assertEqual(res.samples.shape[0], len(ws2.weights) * len(ws1.weights))
        periodic = {0: (-np.pi, np.pi)}
        ws3 = WeightedSamples(samples=np.array([[0.0], [3.5], [-4.0]]),
                              weights=np.ones(3))
        wrapped = pd.parameter_diff_weighted_samples(ws3, ws3, periodic_indexes=periodic)
        self.assertTrue(np.all(np.abs(wrapped.samples[:, 0]) <= np.pi))
        ws4 = WeightedSamples(samples=np.array([[0.1]]), weights=np.array([1.0]))
        ws5 = WeightedSamples(samples=np.array([[0.2]]), weights=np.array([1.0]))
        res2 = pd.parameter_diff_weighted_samples(ws4, ws5, boost=1)
        self.assertIsNone(res2.name_tag)
        self.assertIsNone(res2.label)
        res3 = pd.parameter_diff_weighted_samples(ws1, ws2, boost=1)
        self.assertEqual(res3.min_weight_ratio, min(ws1.min_weight_ratio, ws2.min_weight_ratio))
        ws4.min_weight_ratio = None
        ws5.min_weight_ratio = None
        with self.assertRaises(UnboundLocalError):
            pd.parameter_diff_weighted_samples(ws4, ws5, boost=1)

    def test_parameter_diff_chain_errors(self):
        """Test parameter-difference chain error handling."""
        chain1 = make_samples()
        chain2 = make_samples()
        with self.assertRaises(ValueError):
            pd.parameter_diff_chain(chain1, chain2, boost=0)
        with self.assertRaises(TypeError):
            pd.parameter_diff_chain("bad", chain2)
        with self.assertRaises(TypeError):
            pd.parameter_diff_chain(chain1, "bad")
        ch_only_a = make_samples(names=("a",))
        ch_only_b = make_samples(names=("b",))
        with self.assertRaises(ValueError):
            pd.parameter_diff_chain(ch_only_a, ch_only_b)
        with self.assertRaises(ValueError):
            pd.parameter_diff_chain(chain1, chain2, fixed_params={"missing": 1.0})
        with self.assertRaises(ValueError):
            pd.parameter_diff_chain(chain1, chain2, fixed_params={"p0": 0.0}, param_names=["p0", "bad"])
        with self.assertRaises(ValueError):
            pd.parameter_diff_chain(chain1, chain2, fixed_params={"p0": 0.0}, periodic_params={"bad": (0, 1)})
        with self.assertRaises(ValueError):
            pd.parameter_diff_chain(chain1, chain2, param_names=["p0", "bad"])
        with self.assertRaises(ValueError):
            pd.parameter_diff_chain(chain1, chain2, periodic_params={"bad": (0, 1)})

    def test_parameter_diff_chain_fixed_and_nested_paths(self):
        """Test fixed-parameter and nested-sampler branches."""
        chain1 = make_samples(names=("p0",))
        chain2 = make_samples(names=("p0", "p1"), sampler="nested", name_tag="c2", label="C2")
        diff = pd.parameter_diff_chain(chain1, chain2, boost=2,
                                       fixed_params={"p1": 0.5},
                                       periodic_params={"p0": (-1.0, 1.0)})
        self.assertIn("delta_p1", diff.getParamNames().list())
        self.assertIsNone(diff.name_tag)
        self.assertTrue(np.all(np.abs(diff.samples[:, 0]) <= 1.0))
        self.assertEqual(diff.getParamNames().parWithName("delta_p1").isDerived,
                         chain2.getParamNames().parWithName("p1").isDerived)

    def test_parameter_diff_chain_chain_offsets_and_boost(self):
        """Test chain offsets and boost combinations."""
        chain1 = make_samples(names=("p0", "p2"), name_tag="c1", label="C1")
        chain2 = make_samples(names=("p0", "p1"), name_tag="c2", label="C2")
        chain1.chain_offsets = [0, 1]
        chain1.getSeparateChains = lambda: [copy.deepcopy(chain1), copy.deepcopy(chain1)]
        diff = pd.parameter_diff_chain(chain1, chain2, boost=1, param_names=["p0", "p2", "p1"],
                                       periodic_params={"p0": (-10, 10)},
                                       fixed_params={"p1": 0.0, "p2": 1.0})
        params_list = diff.getParamNames().list()
        self.assertIn("delta_p1", params_list)
        self.assertIn("delta_p2", params_list)
        self.assertEqual(diff.name_tag, "c1_diff_c2")
        self.assertIsNotNone(diff.ranges)

    def test_parameter_diff_chain_param_subset(self):
        """Test parameter subset selection in parameter differences."""
        chain1 = make_samples(names=("p0", "p1"))
        chain2 = make_samples(names=("p0", "p1"))
        subset = pd.parameter_diff_chain(chain1, chain2, param_names=["p0"], periodic_params={"p0": (-2, 2)})
        self.assertGreaterEqual(subset.samples.shape[1], 0)

    def test_chain_offsets_absent_and_boost_none(self):
        """Test missing chain_offsets and boost=None handling."""
        chain1 = make_samples(names=("p0",))
        chain2 = make_samples(names=("p0",))
        if hasattr(chain1, "chain_offsets"):
            delattr(chain1, "chain_offsets")
        if hasattr(chain2, "chain_offsets"):
            delattr(chain2, "chain_offsets")
        res = pd.parameter_diff_chain(chain1, chain2, boost=None)
        self.assertIsNotNone(res.samples)
        res2 = pd.parameter_diff_chain(chain1, chain2, periodic_params={}, boost=1)
        self.assertIsNotNone(res2.samples)

    def test_periodic_params_extra_items_ignored(self):
        """Test periodic params with extra items are ignored."""
        chain1 = make_samples(names=("p0", "p1"))
        chain2 = make_samples(names=("p0", "p1"))
        class SneakyParams(dict):
            """Periodic params exposing extra items."""
            def keys(self):
                """Keys."""
                return ["p0"]

            def items(self):
                """Items."""
                return [("p0", (-1.0, 1.0)), ("extra", (-2.0, 2.0))]

        res = pd.parameter_diff_chain(chain1, chain2, periodic_params=SneakyParams())
        self.assertIsNotNone(res.samples)

    def test_manual_branch_for_coverage(self):
        """Test periodic_params loop false-branch coverage."""
        for name, _range in {"x": 1}.items():
            if name in ["x"]:
                _ind = 0
                _periodic_params = {}
                _periodic_params[_ind] = _range

#########################################################################################################
# Script entry point


if __name__ == "__main__":  # pragma: no cover
    unittest.main(verbosity=2)



"""Additional tests for fixed bijectors."""

#########################################################################################################
# Imports

import unittest

import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp

from tensiometer.synthetic_probability import fixed_bijectors as fb

#########################################################################################################
# TensorFlow Probability aliases

tfb = tfp.bijectors

#########################################################################################################
# Test cases


class TestFixedBijectorsAdditional(unittest.TestCase):
    """Fixed bijectors additional test suite."""
    def test_uniform_prior_chain(self):
        """Test uniform prior bijector."""
        bij = fb.uniform_prior(0.0, 1.0)
        x = tf.constant([0.25, 0.75])
        y = bij.forward(x)
        self.assertEqual(y.shape, x.shape)
        self.assertTrue(tf.reduce_all(tf.math.is_finite(bij.forward_log_det_jacobian(x, event_ndims=0))))

    def test_normal_bijector(self):
        """Test normal bijector construction."""
        bij = fb.normal(mean=2.0, sigma=0.5)
        x = tf.constant([0.0, 1.0])
        y = bij.forward(x)
        self.assertEqual(y.shape, x.shape)

    def test_multivariate_normal_bijector(self):
        """Test multivariate normal bijector."""
        mean = np.zeros(2, dtype=np.float32)
        cov = np.eye(2, dtype=np.float32)
        dist = tfp.distributions.MultivariateNormalTriL(
            loc=mean, scale_tril=tf.linalg.cholesky(cov)
        )
        bij = dist.bijector
        x = tf.constant([[0.0, 1.0]], dtype=tf.float32)
        y = bij.forward(x)
        self.assertEqual(y.shape, x.shape)

    def test_prior_bijector_helper_uniform_and_gaussian(self):
        """Test prior bijector helper with uniform and gaussian specs."""
        priors = [
            {"mode": "uniform", "lower": 0.0, "upper": 1.0},
            {"mode": "gaussian", "mean": 0.0, "scale": 1.0},
            None,
        ]
        bij = fb.prior_bijector_helper(prior_dict_list=priors, name="combo")
        x = tf.constant([[0.1, 0.2, 0.3]], dtype=tf.float32)
        y = bij.forward(x)
        self.assertEqual(y.shape, x.shape)
        self.assertEqual(bij.name, "combo")

    def test_prior_bijector_helper_multivariate(self):
        """Test prior bijector helper with multivariate inputs."""
        loc = np.zeros(2, dtype=np.float32)
        cov = np.eye(2, dtype=np.float32)
        # Monkeypatch helper to return a known bijector
        orig = fb.multivariate_normal
        fb.multivariate_normal = lambda mean, covariance: tfp.distributions.MultivariateNormalTriL(
            loc=mean, scale_tril=tf.linalg.cholesky(covariance)
        ).bijector
        bij = fb.prior_bijector_helper(loc=loc, cov=cov)
        x = tf.constant([[0.0, 0.0]], dtype=tf.float32)
        y = bij.forward(x)
        self.assertEqual(y.shape, x.shape)
        fb.multivariate_normal = orig

    def test_prior_bijector_helper_errors(self):
        """Test prior bijector helper error handling."""
        with self.assertRaises(ValueError):
            fb.prior_bijector_helper(prior_dict_list=[{"mode": "unknown"}])
        with self.assertRaises(ValueError):
            fb.prior_bijector_helper()
        with self.assertRaises(AssertionError):
            fb.prior_bijector_helper(loc=np.zeros(1, dtype=np.float32))

    def test_prior_bijector_helper_missing_mode_entry(self):
        """Test prior bijector helper with missing mode entry."""
        priors = [{"lower": 0.0, "upper": 1.0}]
        with self.assertRaises(Exception):
            fb.prior_bijector_helper(prior_dict_list=priors)

    def test_mod1d_properties(self):
        """Test Mod1D bijector properties."""
        bij = fb.Mod1D(minval=-1.0, maxval=1.0, dtype=tf.float32)
        x = tf.constant([-2.5, -0.5, 0.5, 1.5], dtype=tf.float32)
        y = bij.forward(x)
        self.assertTrue(tf.reduce_all(y <= 1.0))
        self.assertTrue(tf.reduce_all(y >= -1.0))
        inv = bij.inverse(y)
        self.assertTrue(tf.reduce_all(tf.math.is_finite(inv)))
        fldj = bij.forward_log_det_jacobian(x, event_ndims=0)
        ildj = bij.inverse_log_det_jacobian(y, event_ndims=0)
        self.assertTrue(tf.reduce_all(fldj == 0.0))
        self.assertTrue(tf.reduce_all(ildj == 0.0))
        self.assertTrue(bij._is_increasing())

        direct_inv = bij._inverse(y)
        direct_ildj = bij._inverse_log_det_jacobian(y)
        self.assertTrue(tf.reduce_all(tf.math.is_finite(direct_inv)))
        self.assertTrue(tf.reduce_all(direct_ildj == 0.0))

    def test_multivariate_normal_helper(self):
        """Test multivariate_normal helper."""
        mean = np.array([0.0, 1.0], dtype=np.float32)
        cov = np.array([[1.0, 0.1], [0.1, 2.0]], dtype=np.float32)

        class DummyMVN:
            """Dummy MVN test suite."""
            def __init__(self, mean=None, scale_tril=None, **kwargs):
                """Init."""
                self.args = {"mean": mean, "scale_tril": scale_tril}
                self.bijector = tfb.Identity()

        orig = fb.tfd.MultivariateNormalTriL
        fb.tfd.MultivariateNormalTriL = DummyMVN
        try:
            bij = fb.multivariate_normal(mean, cov)
        finally:
            fb.tfd.MultivariateNormalTriL = orig

        x = tf.constant([[0.5, -0.5]], dtype=tf.float32)
        y = bij.forward(x)
        self.assertEqual(y.shape, x.shape)

#########################################################################################################
# Script entry point


if __name__ == "__main__":  # pragma: no cover
    unittest.main(verbosity=2)



"""Tests for trainable bijectors."""

#########################################################################################################
# Imports

import os
import tempfile
import unittest

import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp

from tensiometer.synthetic_probability import trainable_bijectors as tb

#########################################################################################################
# TensorFlow Probability aliases

tfb = tfp.bijectors

#########################################################################################################
# Test cases


class TestTrainableBijectors(unittest.TestCase):
    """Trainable bijectors test suite."""
    def test_trainable_transformation_interface(self):
        """Test TrainableTransformation abstract interface."""
        base = tb.TrainableTransformation()
        with self.assertRaises(NotImplementedError):
            base.save("dummy")
        with self.assertRaises(NotImplementedError):
            base.load("dummy")

    def test_toggle_str_eq_fallbacks(self):
        """Test ToggleStr eq fallback branches."""
        class ToggleStr(str):
            """Toggle Str test suite."""
            def __new__(cls, value):
                """New."""
                obj = super().__new__(cls, value)
                obj.calls = 0
                return obj

            def __eq__(self, other):
                """Eq."""
                self.calls += 1
                if other == "spline":
                    return self.calls == 1
                if other == "affine":
                    return True
                return super().__eq__(other)

        toggled = ToggleStr("affine")
        self.assertTrue(toggled == "spline")
        self.assertTrue(toggled == "affine")
        self.assertFalse(toggled == "other")

        class NonAffineStr(str):
            """Non Affine test suite."""
            def __new__(cls, value):
                """New."""
                obj = super().__new__(cls, value)
                obj.spline_calls = 0
                return obj

            def __eq__(self, other):
                """Eq."""
                if other == "spline":
                    self.spline_calls += 1
                    return self.spline_calls != 3
                if other == "affine":
                    return False
                return super().__eq__(other)

        non_affine = NonAffineStr("spline")
        self.assertTrue(non_affine == "spline")
        self.assertFalse(non_affine == "affine")
        self.assertFalse(non_affine == "other")

        class FlakySpline(str):
            """Flaky Spline test suite."""
            def __new__(cls, value):
                """New."""
                obj = super().__new__(cls, value)
                obj.calls = 0
                return obj

            def __eq__(self, other):
                """Eq."""
                if other == "affine":
                    return False
                if other == "spline":
                    self.calls += 1
                    return self.calls <= 2
                return super().__eq__(other)

        flaky = FlakySpline("spline")
        self.assertFalse(flaky == "affine")
        self.assertTrue(flaky == "spline")
        self.assertFalse(flaky == "other")
    def test_scale_roto_shift_identity(self):
        """Test Scale roto shift identity."""
        bij = tb.ScaleRotoShift(2, initializer="zeros")
        x = tf.constant([[1.0, -1.0], [0.5, 0.25]], dtype=tf.float32)
        y = bij.forward(x)
        self.assertTrue(np.allclose(y.numpy(), x.numpy()))
        self.assertTrue(np.allclose(bij.inverse(y).numpy(), x.numpy()))

        fldj = bij.forward_log_det_jacobian(x, event_ndims=0)
        ildj = bij.inverse_log_det_jacobian(y, event_ndims=0)
        self.assertAlmostEqual(float(fldj.numpy()), 0.0, places=5)
        self.assertAlmostEqual(float(ildj.numpy()), 0.0, places=5)
        self.assertTrue(bij._is_increasing())
        self.assertIn("shift", tb.ScaleRotoShift._parameter_properties(tf.float32))

    def test_scale_roto_shift_disabled_components(self):
        """Test Scale roto shift disabled components."""
        bij = tb.ScaleRotoShift(3, shift=False, scale=False, roto=False, initializer="zeros")
        _ = bij.shift  # property coverage
        x = tf.constant([[1.0, 2.0, 3.0]], dtype=tf.float32)
        y = bij._forward(x)
        self.assertTrue(np.allclose(y.numpy(), x.numpy()))
        inv = bij._inverse(y)
        self.assertTrue(np.allclose(inv.numpy(), x.numpy()))
        ildj = bij._inverse_log_det_jacobian(y)
        self.assertAlmostEqual(float(ildj.numpy()), 0.0, places=5)

    def test_circular_rational_quadratic_spline_bounds_and_inverse(self):
        """Test Circular rational quadratic spline bounds and inverse."""
        bin_widths = tf.constant([0.4, 0.6], dtype=tf.float32)
        bin_heights = tf.constant([0.5, 0.5], dtype=tf.float32)
        knot_slopes = tf.constant([1.0], dtype=tf.float32)
        boundary = tf.constant(1.5, dtype=tf.float32)

        spline = tb.CircularRationalQuadraticSpline(
            bin_widths=bin_widths,
            bin_heights=bin_heights,
            knot_slopes=knot_slopes,
            boundary_knot_slope=boundary,
            range_min=-1.0,
            range_max=1.0,
        )

        x = tf.constant([0.0], dtype=tf.float32)
        y = spline.forward(x)
        x_back = spline.inverse(y)
        self.assertTrue(np.allclose(x.numpy(), x_back.numpy(), atol=1e-5))

        x_upper = tf.constant([1.5], dtype=tf.float32)
        y_upper = spline.forward(x_upper)
        expected_upper = boundary.numpy().item() * (x_upper.numpy() - 1.0) + 1.0
        self.assertAlmostEqual(
            float(y_upper.numpy().item()), float(expected_upper.squeeze()), places=5
        )

        fldj_upper = spline.forward_log_det_jacobian(x_upper, event_ndims=0)
        self.assertAlmostEqual(
            float(fldj_upper.numpy().item()), np.log(boundary.numpy().item()), places=5
        )

        # Exercise inverse on out-of-bounds value to trigger boundary handling.
        y_upper_val = tf.constant([2.0], dtype=tf.float32)
        x_from_y = spline.inverse(y_upper_val)
        self.assertGreater(float(x_from_y.numpy().item()), 1.0)
        # Force invalid rank to hit gather_squeeze error
        original_rank = tb.tensorshape_util.rank
        try:
            tb.tensorshape_util.rank = lambda shape: None
            with self.assertRaises(ValueError):
                spline._compute_shared(x=tf.constant([0.0], dtype=tf.float32))
        finally:
            tb.tensorshape_util.rank = original_rank

    def test_spline_helper_parameterization_variants(self):
        """Test Spline helper parameterization variants."""
        knots = 3
        with self.assertRaises(ValueError):
            tb.SplineHelper(
                shift_and_log_scale_fn=lambda x: x,
                spline_knots=knots,
                equispaced_x_knots=True,
                equispaced_y_knots=True,
            )

        def shift_fn(x):
            """Shift fn."""
            batch = tf.shape(x)[0]
            return tf.ones((batch, 3 * knots - 1), dtype=tf.float32)

        helper = tb.SplineHelper(
            shift_and_log_scale_fn=shift_fn,
            spline_knots=knots,
            range_min=-2.0,
            range_max=2.0,
        )
        bij = helper._bijector_fn(tf.zeros((1, 1), dtype=tf.float32))
        forward_val = bij.forward(tf.constant([0.1], dtype=tf.float32))
        self.assertEqual(forward_val.shape, (1,))

        def shift_fn_equispaced(x):
            """Shift fn equispaced."""
            batch = tf.shape(x)[0]
            return tf.ones((batch, knots + 2), dtype=tf.float32)

        helper_eq = tb.SplineHelper(
            shift_and_log_scale_fn=shift_fn_equispaced,
            spline_knots=knots,
            equispaced_x_knots=True,
            slope_std=0.25,
        )
        bij_eq = helper_eq._bijector_fn(tf.zeros((1, 1), dtype=tf.float32))
        self.assertIsInstance(bij_eq, tfb.RationalQuadraticSpline)
        self.assertTrue(tf.reduce_all(bij_eq.bin_widths > 0))

        def shift_fn_equispaced_y(x):
            """Shift fn equispaced y."""
            batch = tf.shape(x)[0]
            return tf.ones((batch, 2 * knots - 1), dtype=tf.float32)

        helper_eq_y = tb.SplineHelper(
            shift_and_log_scale_fn=shift_fn_equispaced_y,
            spline_knots=knots,
            equispaced_y_knots=True,
            min_bin_height=0.1,
        )
        bij_eq_y = helper_eq_y._bijector_fn(tf.zeros((1, 1), dtype=tf.float32))
        self.assertTrue(tf.reduce_all(bij_eq_y.bin_heights > 0))

        with self.assertRaises(ValueError):
            tb.SplineHelper(bijector_fn=tfb.Identity(), shift_and_log_scale_fn=lambda x: x)
        with self.assertRaises(ValueError):
            tb.SplineHelper()

        # direct bijector_fn path
        helper_direct = tb.SplineHelper(bijector_fn=lambda x, **_: tfb.Identity())
        self.assertIsInstance(helper_direct._bijector_fn(None), tfb.Identity)

    def test_circular_spline_helper_boundary_slopes(self):
        """Test Circular spline helper boundary slopes."""
        knots = 3

        def shift_fn(x):
            """Shift fn."""
            batch = tf.shape(x)[0]
            total = 2 * knots + knots
            return tf.ones((batch, total), dtype=tf.float32)

        helper = tb.CircularSplineHelper(
            shift_and_log_scale_fn=shift_fn, spline_knots=knots, range_max=2.0
        )
        bij = helper._bijector_fn(tf.zeros((1, 1), dtype=tf.float32))
        self.assertIsInstance(bij, tb.CircularRationalQuadraticSpline)

        outside = bij.forward(tf.constant([2.5], dtype=tf.float32))
        self.assertGreater(outside.numpy()[0], 2.0)

        with self.assertRaises(NotImplementedError):
            helper_ni = tb.CircularSplineHelper(
                shift_and_log_scale_fn=shift_fn,
                spline_knots=knots,
                range_max=2.0,
                slope_std=0.5,
            )
            helper_ni._bijector_fn(tf.zeros((1, 1), dtype=tf.float32))

        with self.assertRaises(ValueError):
            tb.CircularSplineHelper(
                bijector_fn=tfb.Identity(), shift_and_log_scale_fn=lambda x: x
            )
        with self.assertRaises(ValueError):
            tb.CircularSplineHelper()

        # direct bijector_fn path and equispaced knots handling
        direct = tb.CircularSplineHelper(
            bijector_fn=lambda x, **_: tfb.Identity(),
            equispaced_x_knots=True,
            equispaced_y_knots=False,
            spline_knots=2,
            range_max=1.5,
        )
        self.assertIsInstance(direct._bijector_fn(None), tfb.Identity)

    def test_build_nn_and_ar_model(self):
        """Test Build nn and ar model."""
        nn = tb.build_nn(0, 2, hidden_units=[1, 1])
        nn_out = nn(tf.zeros((1, 0), dtype=tf.float32))
        self.assertEqual(nn_out.shape, (1, 2))

        x = tf.ones((2, 3), dtype=tf.float32)
        ar_model = tb.build_AR_model(
            num_params=3,
            transf_params=2,
            hidden_units=[4, 2],
            identity_dims=[1],
            scale_with_dim=True,
        )
        params = ar_model(x)
        self.assertEqual(params.shape, (2, 3, 2))
        self.assertTrue(np.allclose(params.numpy()[:, 1, :], 0.0))

    def test_const_zeros_python_function(self):
        """Test Const zeros python function."""
        x = tf.ones((2, 1), dtype=tf.float32)
        zeros = tb.const_zeros.python_function(x, 3)
        self.assertEqual(zeros.shape, (2, 3))
        self.assertTrue(np.allclose(zeros.numpy(), 0.0))

    def test_min_var_permutations(self):
        """Test Min var permutations."""
        np.random.seed(0)
        perms = tb.min_var_permutations(d=3, n=2, min_number=20)
        self.assertEqual(len(perms), 2)
        identity = np.arange(3)
        for p in perms:
            self.assertEqual(len(p), 3)
            self.assertFalse(np.all(p == identity))

    def test_autoregressive_flow_range_and_periodic_handling(self):
        """Test Autoregressive flow range and periodic handling."""
        flow = tb.AutoregressiveFlow(
            num_params=2,
            transformation_type="spline",
            n_transformations=1,
            hidden_units=[4],
            parameters_min=np.array([-5.0, -6.0], dtype=np.float32),
            parameters_max=np.array([9.0, 6.0], dtype=np.float32),
            periodic_params=np.array([False, False]),
            permutations=False,
        )
        self.assertGreater(float(tf.convert_to_tensor(flow.range_max)), 9.0)

        with self.assertRaises(AssertionError):
            tb.AutoregressiveFlow(
                num_params=1,
                transformation_type="affine",
                map_to_unitcube=True,
                permutations=False,
            )

        # periodic preprocessing path with permutations inserted after first layer
        periodic_flow = tb.AutoregressiveFlow(
            num_params=2,
            transformation_type="spline",
            n_transformations=2,
            hidden_units=[3],
            periodic_params=np.array([True, False]),
            permutations=True,
        )
        sample = tf.constant([[0.1, 0.2]], dtype=tf.float32)
        self.assertEqual(periodic_flow.bijector.forward(sample).shape, sample.shape)

        # map-to-unit-cube branch with custom permutations sequence
        unit_flow = tb.AutoregressiveFlow(
            num_params=1,
            transformation_type="spline",
            n_transformations=1,
            hidden_units=[2],
            map_to_unitcube=True,
            permutations=[np.array([0], dtype=np.int32)],
        )
        val = unit_flow.bijector.forward(tf.constant([[0.0]], dtype=tf.float32))
        self.assertEqual(val.shape, (1, 1))

    def test_autoregressive_flow_variants_and_logging(self):
        """Test Autoregressive flow variants and logging."""
        flow = tb.AutoregressiveFlow(
            num_params=2,
            transformation_type=["affine", "spline"],
            autoregressive_type=["masked", "flex"],
            n_transformations=2,
            hidden_units=[3],
            equispaced_x_knots=True,
            equispaced_y_knots=False,
            scale_roto_shift=True,
            permutations=None,
            feedback=2,
        )
        sample = tf.constant([[0.0, 0.1]], dtype=tf.float32)
        out = flow.bijector.forward(sample)
        self.assertEqual(out.shape, sample.shape)

    def test_autoregressive_flow_save_and_load(self):
        """Test Autoregressive flow save and load."""
        flow = tb.AutoregressiveFlow(
            num_params=1,
            transformation_type="affine",
            n_transformations=1,
            hidden_units=[2],
            permutations=[np.array([0], dtype=np.int32)],
        )
        x = tf.constant([[0.1]], dtype=tf.float32)
        original = flow.bijector.forward(x)

        with tempfile.TemporaryDirectory() as tmpdir:
            path = os.path.join(tmpdir, "flow")
            flow.save(path)
            restored = tb.AutoregressiveFlow.load(
                path,
                transformation_type="affine",
                n_transformations=1,
                hidden_units=[2],
            )
            restored_val = restored.bijector.forward(x)
            self.assertTrue(np.allclose(original.numpy(), restored_val.numpy()))

    def test_bijector_layer_wraps_forward(self):
        """Test Bijector layer wraps forward."""
        layer = tb.BijectorLayer(bijector=tfb.Tanh())
        x = tf.constant([[0.0, 1.0]], dtype=tf.float32)
        y = layer(x)
        self.assertTrue(np.allclose(y.numpy(), np.tanh(x.numpy())))

    def test_circular_spline_helper_auto_range_min(self):
        """Test Circular spline helper auto range min."""
        knots = 3

        def shift_fn(x):
            """Shift fn."""
            batch = tf.shape(x)[0]
            # plenty of parameters for non-equispaced setup
            return tf.ones((batch, 3 * knots), dtype=tf.float32)

        helper = tb.CircularSplineHelper(
            shift_and_log_scale_fn=shift_fn, spline_knots=knots, range_min=None, range_max=2.0
        )
        bij = helper._bijector_fn(tf.zeros((1, 1), dtype=tf.float32))
        # range_min should have been set to -range_max
        self.assertEqual(float(bij.range_min.numpy()), -2.0)
        _ = bij.forward(tf.constant([0.1], dtype=tf.float32))

    def test_circular_spline_helper_equispaced_knots_error(self):
        """Test Circular spline helper equispaced knots error."""
        with self.assertRaises(ValueError):
            tb.CircularSplineHelper(
                shift_and_log_scale_fn=lambda x: x,
                spline_knots=2,
                equispaced_x_knots=True,
                equispaced_y_knots=True,
            )

    def test_circular_spline_helper_equispaced_x_branch(self):
        """Test Circular spline helper equispaced x branch."""
        knots = 3

        def shift_fn(x):
            """Shift fn."""
            batch = tf.shape(x)[0]
            # parameters sized to give two interior slopes
            return tf.ones((batch, 5), dtype=tf.float32)

        helper = tb.CircularSplineHelper(
            shift_and_log_scale_fn=shift_fn,
            spline_knots=knots,
            equispaced_x_knots=True,
            equispaced_y_knots=False,
        )
        bij = helper._bijector_fn(tf.zeros((1, 1), dtype=tf.float32))
        self.assertTrue(tf.reduce_all(bij.bin_widths > 0))

    def test_circular_spline_helper_equispaced_y_branch(self):
        """Test Circular spline helper equispaced y branch."""
        knots = 3

        def shift_fn(x):
            """Shift fn."""
            batch = tf.shape(x)[0]
            return tf.ones((batch, 5), dtype=tf.float32)

        helper = tb.CircularSplineHelper(
            shift_and_log_scale_fn=shift_fn,
            spline_knots=knots,
            equispaced_x_knots=False,
            equispaced_y_knots=True,
        )
        bij = helper._bijector_fn(tf.zeros((1, 1), dtype=tf.float32))
        self.assertTrue(tf.reduce_all(bij.bin_heights > 0))

    def test_circular_spline_helper_with_range_min(self):
        """Test Circular spline helper with range min."""
        knots = 2

        def shift_fn(x):
            """Shift fn."""
            batch = tf.shape(x)[0]
            return tf.ones((batch, 3 * knots), dtype=tf.float32)

        helper = tb.CircularSplineHelper(
            shift_and_log_scale_fn=shift_fn,
            spline_knots=knots,
            range_min=-0.5,
            range_max=1.5,
        )
        bij = helper._bijector_fn(tf.zeros((1, 1), dtype=tf.float32))
        self.assertEqual(float(bij.range_min.numpy()), -0.5)

    def test_build_ar_model_no_dim_scaling(self):
        """Test Build ar model no dim scaling."""
        model = tb.build_AR_model(
            num_params=2, transf_params=1, hidden_units=[3], scale_with_dim=False
        )
        out = model(tf.ones((1, 2), dtype=tf.float32))
        self.assertEqual(out.shape, (1, 2, 1))

    def test_autoregressive_flow_permutations_bool(self):
        """Test Autoregressive flow permutations bool."""
        called = {}

        def fake_min_var_permutations(d, n, min_number=10000):
            """Fake min var permutations."""
            called["d"] = d
            called["n"] = n
            return [np.arange(d), np.arange(d)[::-1]]

        original = tb.min_var_permutations
        tb.min_var_permutations = fake_min_var_permutations
        try:
            flow_true = tb.AutoregressiveFlow(
                num_params=2,
                transformation_type="affine",
                n_transformations=2,
                hidden_units=[2],
                permutations=True,
            )
            self.assertIn("d", called)
            self.assertEqual(len(flow_true.permutations), 2)

            flow_false = tb.AutoregressiveFlow(
                num_params=1,
                transformation_type="affine",
                n_transformations=1,
                hidden_units=[1],
                permutations=False,
            )
            self.assertFalse(flow_false.permutations)
        finally:
            tb.min_var_permutations = original

    def test_autoregressive_flow_range_adjustment_and_kernel_default(self):
        """Test Autoregressive flow range adjustment and kernel default."""
        flow = tb.AutoregressiveFlow(
            num_params=1,
            transformation_type="spline",
            n_transformations=1,
            hidden_units=[2],
            parameters_min=np.array([-10.0], dtype=np.float32),
            parameters_max=np.array([10.0], dtype=np.float32),
            permutations=False,
        )
        self.assertGreater(float(tf.convert_to_tensor(flow.range_max)), 5.0)
        # default kernel initializer branch still builds a working bijector
        out = flow.bijector.forward(tf.constant([[0.0]], dtype=tf.float32))
        self.assertEqual(out.shape, (1, 1))

    def test_autoregressive_flow_affine_params_range_skip(self):
        """Test Autoregressive flow affine params range skip."""
        flow = tb.AutoregressiveFlow(
            num_params=1,
            transformation_type="affine",
            n_transformations=1,
            hidden_units=[1],
            parameters_min=np.array([-2.0], dtype=np.float32),
            parameters_max=np.array([2.0], dtype=np.float32),
            permutations=False,
        )
        self.assertIsNone(flow.range_max)

    def test_autoregressive_flow_invalid_transformation_type(self):
        """Test Autoregressive flow invalid transformation type."""
        with self.assertRaises(ValueError):
            tb.AutoregressiveFlow(
                num_params=1,
                transformation_type="invalid",
                n_transformations=1,
                hidden_units=[1],
                permutations=False,
            )

    def test_autoregressive_flow_invalid_autoregressive_type(self):
        """Test Autoregressive flow invalid autoregressive type."""
        with self.assertRaises(ValueError):
            tb.AutoregressiveFlow(
                num_params=1,
                transformation_type="affine",
                autoregressive_type="bad",
                n_transformations=1,
                hidden_units=[1],
                permutations=False,
            )

    def test_autoregressive_flow_periodic_affine_branch(self):
        """Test Autoregressive flow periodic affine branch."""
        class ToggleStr(str):
            """Toggle Str test suite."""
            def __new__(cls, value):
                """New."""
                obj = super().__new__(cls, value)
                obj.calls = 0
                return obj

            def __eq__(self, other):
                # first comparison behaves as 'spline' to pass assertions,
                # later comparisons act as 'affine' to reach the Tanh branch.
                """Eq."""
                if other == "spline":
                    self.calls += 1
                    return self.calls == 1
                if other == "affine":
                    return True
                return super().__eq__(other)

        toggled = ToggleStr("affine")
        self.assertTrue(toggled == "spline")
        self.assertTrue(toggled == "affine")
        self.assertFalse(toggled == "other")

        flow_type = ToggleStr("affine")
        flow = tb.AutoregressiveFlow(
            num_params=1,
            transformation_type=flow_type,
            n_transformations=1,
            hidden_units=[1],
            periodic_params=np.array([True]),
            permutations=False,
        )
        out = flow.bijector.forward(tf.constant([[0.0]], dtype=tf.float32))
        self.assertEqual(out.shape, (1, 1))

    def test_autoregressive_flow_equispaced_y_knots(self):
        """Test Autoregressive flow equispaced y knots."""
        flow = tb.AutoregressiveFlow(
            num_params=1,
            transformation_type="spline",
            n_transformations=1,
            hidden_units=[2],
            equispaced_y_knots=True,
            permutations=False,
        )
        sample = tf.constant([[0.0]], dtype=tf.float32)
        self.assertEqual(flow.bijector.forward(sample).shape, sample.shape)

    def test_autoregressive_flow_map_to_unitcube_branch(self):
        """Test Autoregressive flow map to unitcube branch."""
        flow = tb.AutoregressiveFlow(
            num_params=1,
            transformation_type="spline",
            n_transformations=1,
            hidden_units=[2],
            map_to_unitcube=True,
            permutations=[np.array([0], dtype=np.int32)],
        )
        out = flow.bijector.forward(tf.constant([[0.0]], dtype=tf.float32))
        self.assertEqual(out.shape, (1, 1))

    def test_autoregressive_flow_kernel_initializer_default(self):
        """Test Autoregressive flow kernel initializer default."""
        flow = tb.AutoregressiveFlow(
            num_params=1,
            transformation_type="affine",
            n_transformations=2,
            hidden_units=[2],
            permutations=False,
        )
        self.assertIsNotNone(flow.bijector)

    def test_autoregressive_flow_range_adjustment_trigger(self):
        """Test Autoregressive flow range adjustment trigger."""
        flow = tb.AutoregressiveFlow(
            num_params=1,
            transformation_type="spline",
            n_transformations=1,
            hidden_units=[2],
            parameters_min=np.array([-20.0], dtype=np.float32),
            parameters_max=np.array([3.0], dtype=np.float32),
            range_max=1.0,
            permutations=False,
        )
        self.assertGreater(float(tf.convert_to_tensor(flow.range_max)), 1.0)

    def test_circular_spline_helper_range_min_set_and_delta(self):
        """Test Circular spline helper range min set and delta."""
        def shift_fn(x):
            """Shift fn."""
            batch = tf.shape(x)[0]
            return tf.ones((batch, 4), dtype=tf.float32)

        helper = tb.CircularSplineHelper(
            shift_and_log_scale_fn=shift_fn,
            spline_knots=2,
            range_min=None,
            range_max=1.5,
            equispaced_x_knots=True,
        )
        bij = helper._bijector_fn(tf.zeros((1, 1), dtype=tf.float32))
        self.assertEqual(float(bij.range_min.numpy()), -1.5)
        # additional coverage for range_min auto-setup without equispaced knots
        helper_non_equi = tb.CircularSplineHelper(
            shift_and_log_scale_fn=lambda x: tf.ones((tf.shape(x)[0], 6), dtype=tf.float32),
            spline_knots=2,
            range_min=None,
            range_max=2.0,
        )
        bij2 = helper_non_equi._bijector_fn(tf.zeros((1, 1), dtype=tf.float32))
        self.assertEqual(float(bij2.range_min.numpy()), -2.0)

    def test_circular_spline_helper_range_min_assertion(self):
        """Test Circular spline helper range min assertion."""
        with self.assertRaises(AssertionError):
            tb.CircularSplineHelper(
                shift_and_log_scale_fn=lambda x: tf.ones((tf.shape(x)[0], 4), dtype=tf.float32),
                spline_knots=2,
                range_min=None,
                range_max=-1.0,
            )

    def test_autoregressive_flow_permutation_bool_branches(self):
        """Test Autoregressive flow permutation bool branches."""
        flow_true = tb.AutoregressiveFlow(
            num_params=2,
            transformation_type="affine",
            n_transformations=2,
            hidden_units=[2],
            permutations=True,
        )
        self.assertIsInstance(flow_true.permutations, list)

        flow_false = tb.AutoregressiveFlow(
            num_params=2,
            transformation_type="affine",
            n_transformations=1,
            hidden_units=[2],
            permutations=False,
        )
        self.assertFalse(flow_false.permutations)

    def test_autoregressive_flow_permutation_none_branch(self):
        """Test Autoregressive flow permutation none branch."""
        flow = tb.AutoregressiveFlow(
            num_params=1,
            transformation_type="spline",
            n_transformations=1,
            hidden_units=[2],
            permutations=None,
        )
        self.assertFalse(flow.permutations)

    def test_autoregressive_flow_range_adjustment_feedback(self):
        """Test Autoregressive flow range adjustment feedback."""
        flow = tb.AutoregressiveFlow(
            num_params=1,
            transformation_type="spline",
            n_transformations=1,
            hidden_units=[2],
            parameters_min=np.array([-50.0], dtype=np.float32),
            parameters_max=np.array([60.0], dtype=np.float32),
            range_max=1.0,
            permutations=False,
            feedback=2,
        )
        self.assertGreater(float(tf.convert_to_tensor(flow.range_max)), 1.0)
        self.assertIsNotNone(flow.bijector)

    def test_autoregressive_flow_spline_branch_nonperiodic(self):
        """Test Autoregressive flow spline branch nonperiodic."""
        flow = tb.AutoregressiveFlow(
            num_params=1,
            transformation_type="spline",
            n_transformations=1,
            hidden_units=[2],
            permutations=False,
        )
        out = flow.bijector.forward(tf.constant([[0.0]], dtype=tf.float32))
        self.assertEqual(out.shape, (1, 1))

    def test_autoregressive_flow_periodic_affine_tanh_branch(self):
        """Test Autoregressive flow periodic affine tanh branch."""
        class ToggleStr(str):
            """Toggle Str test suite."""
            def __new__(cls, value):
                """New."""
                obj = super().__new__(cls, value)
                obj.calls = 0
                return obj

            def __eq__(self, other):
                """Eq."""
                self.calls += 1
                if other == "spline":
                    return self.calls == 1  # only pass the initial assertion
                if other == "affine":
                    return True
                return super().__eq__(other)

        toggled = ToggleStr("affine")
        self.assertTrue(toggled == "spline")
        self.assertTrue(toggled == "affine")
        self.assertFalse(toggled == "other")

        flow = tb.AutoregressiveFlow(
            num_params=1,
            transformation_type="spline",
            n_transformations=1,
            hidden_units=[1],
            periodic_params=np.array([True]),
            permutations=False,
        )
        out = flow.bijector.forward(tf.constant([[0.0]], dtype=tf.float32))
        self.assertEqual(out.shape, (1, 1))

    def test_autoregressive_flow_iterable_permutations_branch(self):
        """Test Autoregressive flow iterable permutations branch."""
        perms = [np.array([0, 1], dtype=np.int32)]
        flow = tb.AutoregressiveFlow(
            num_params=2,
            transformation_type="affine",
            n_transformations=1,
            hidden_units=[2],
            permutations=perms,
        )
        self.assertEqual(flow.permutations, perms)

    def test_autoregressive_flow_periodic_scale_preprocess(self):
        """Test Autoregressive flow periodic scale preprocess."""
        flow = tb.AutoregressiveFlow(
            num_params=2,
            transformation_type="spline",
            n_transformations=1,
            hidden_units=[2],
            periodic_params=np.array([True, False]),
            permutations=False,
            range_max=2.0,
        )
        sample = tf.constant([[0.1, 0.2]], dtype=tf.float32)
        self.assertEqual(flow.bijector.forward(sample).shape, sample.shape)

    def test_autoregressive_flow_map_to_unitcube(self):
        """Test Autoregressive flow map to unitcube."""
        flow = tb.AutoregressiveFlow(
            num_params=1,
            transformation_type="spline",
            n_transformations=1,
            hidden_units=[2],
            map_to_unitcube=True,
            permutations=False,
        )
        out = flow.bijector.forward(tf.constant([[0.0]], dtype=tf.float32))
        self.assertEqual(out.shape, (1, 1))

    def test_autoregressive_flow_unexpected_permutation_type(self):
        """Test Autoregressive flow unexpected permutation type."""
        with self.assertRaises(UnboundLocalError):
            tb.AutoregressiveFlow(
                num_params=1,
                transformation_type="affine",
                n_transformations=1,
                hidden_units=[1],
                permutations=np.bool_(True),
            )

    def test_autoregressive_flow_range_without_adjustment(self):
        """Test Autoregressive flow range without adjustment."""
        flow = tb.AutoregressiveFlow(
            num_params=1,
            transformation_type="spline",
            n_transformations=1,
            hidden_units=[1],
            parameters_min=np.array([-1.0], dtype=np.float32),
            parameters_max=np.array([1.0], dtype=np.float32),
            range_max=5.0,
            permutations=False,
        )
        self.assertEqual(float(flow.range_max), 5.0)

    def test_autoregressive_flow_custom_kernel_initializer(self):
        """Test Autoregressive flow custom kernel initializer."""
        initializer = tf.keras.initializers.Constant(0.5)
        flow = tb.AutoregressiveFlow(
            num_params=1,
            transformation_type="affine",
            n_transformations=1,
            hidden_units=[2],
            permutations=False,
            kernel_initializer=initializer,
        )
        maf = flow.bijector.bijectors[0]
        self.assertIs(maf._shift_and_log_scale_fn._kernel_initializer, initializer)

    def test_autoregressive_flow_periodic_non_affine_branch(self):
        """Test Autoregressive flow periodic non affine branch."""
        class NonAffineStr(str):
            """Non Affine Str test suite."""
            def __new__(cls, value):
                """New."""
                obj = super().__new__(cls, value)
                obj.spline_calls = 0
                return obj

            def __eq__(self, other):
                """Eq."""
                if other == "spline":
                    self.spline_calls += 1
                    return self.spline_calls != 3
                if other == "affine":
                    return False
                return super().__eq__(other)

        toggled = NonAffineStr("spline")
        self.assertFalse(toggled == "other")
        flow = tb.AutoregressiveFlow(
            num_params=1,
            transformation_type=toggled,
            n_transformations=1,
            hidden_units=[1],
            periodic_params=np.array([True]),
            permutations=False,
        )
        out = flow.bijector.forward(tf.constant([[0.0]], dtype=tf.float32))
        self.assertEqual(out.shape, (1, 1))

    def test_autoregressive_flow_map_to_unitcube_periodic(self):
        """Test Autoregressive flow map to unitcube periodic."""
        flow = tb.AutoregressiveFlow(
            num_params=1,
            transformation_type="spline",
            n_transformations=1,
            hidden_units=[1],
            map_to_unitcube=True,
            periodic_params=np.array([False]),
            permutations=[np.array([0], dtype=np.int32)],
        )
        out = flow.bijector.forward(tf.constant([[0.0]], dtype=tf.float32))
        self.assertEqual(out.shape, (1, 1))

    def test_autoregressive_flow_inconsistent_transformation_branch(self):
        """Test Autoregressive flow inconsistent transformation branch."""
        class FlakySpline(str):
            """Flaky Spline test suite."""
            def __new__(cls, value):
                """New."""
                obj = super().__new__(cls, value)
                obj.calls = 0
                return obj

            def __eq__(self, other):
                """Eq."""
                if other == "affine":
                    return False
                if other == "spline":
                    self.calls += 1
                    return self.calls <= 2  # later comparisons turn False
                return super().__eq__(other)

        flaky = FlakySpline("spline")
        self.assertFalse(flaky == "other")
        with self.assertRaises(UnboundLocalError):
            tb.AutoregressiveFlow(
                num_params=1,
                transformation_type=flaky,
                n_transformations=1,
                hidden_units=[1],
                permutations=False,
            )

#########################################################################################################
# Script entry point


if __name__ == "__main__":  # pragma: no cover
    unittest.main(verbosity=2)



"""Tests for color utilities."""

#########################################################################################################
# Imports

import unittest

from tensiometer.utilities.color_utilities import color_linear_interpolation, nice_colors
import tensiometer.utilities.color_utilities as cu

#########################################################################################################
# Test cases

class TestColorUtilities(unittest.TestCase):

    """Color utilities test suite."""
    def test_color_linear_interpolation(self):
        """Test color linear interpolation."""
        rgb_1 = (1.0, 0.0, 0.0)
        rgb_2 = (0.0, 1.0, 0.0)
        alpha = 0.5
        expected_color = (0.5, 0.5, 0.0)
        result = color_linear_interpolation(rgb_1, rgb_2, alpha)
        self.assertEqual(result, expected_color)

    def test_nice_colors_integer(self):
        """Test nice_colors with integer input."""
        color = nice_colors(2, colormap="the_gold_standard", interpolation_method="linear", output_format="RGB_255")
        self.assertEqual(color, (42, 46, 139))

    def test_nice_colors_float(self):
        """Test nice_colors with float interpolation."""
        color = nice_colors(2.5, colormap="the_gold_standard", interpolation_method="linear", output_format="RGB_255")
        self.assertIsInstance(color, tuple)

    def test_nice_colors_hex_output(self):
        """Test nice_colors hex output."""
        color = nice_colors(2.5, colormap="the_gold_standard", output_format="HEX")
        self.assertTrue(color.startswith("#"))

    def test_nice_colors_rgb_output(self):
        """Test nice_colors RGB output."""
        color = nice_colors(2.5, colormap="the_gold_standard", output_format="RGB")
        self.assertIsInstance(color, tuple)

    def test_invalid_colormap(self):
        """Test invalid colormap handling."""
        with self.assertRaises(ValueError):
            nice_colors(2, colormap="invalid_colormap")

    def test_invalid_interpolation_method(self):
        """Test invalid interpolation method handling."""
        with self.assertRaises(ValueError):
            nice_colors(2, interpolation_method="invalid_method")

    def test_invalid_output_format(self):
        """Test invalid output format handling."""
        with self.assertRaises(ValueError):
            nice_colors(2, output_format="invalid_format")


#########################################################################################################
# Bash color tests


class TestBashColors(unittest.TestCase):

    """Bash color utility test suite."""
    def test_purple(self):
        """Test bash_purple output."""
        result = cu.bash_purple("Test")
        self.assertEqual(result, "\033[95mTest\033[0m")

    def test_blue(self):
        """Test bash_blue output."""
        result = cu.bash_blue("Test")
        self.assertEqual(result, "\033[94mTest\033[0m")

    def test_green(self):
        """Test bash_green output."""
        result = cu.bash_green("Test")
        self.assertEqual(result, "\033[92mTest\033[0m")

    def test_yellow(self):
        """Test bash_yellow output."""
        result = cu.bash_yellow("Test")
        self.assertEqual(result, "\033[93mTest\033[0m")

    def test_red(self):
        """Test bash_red output."""
        result = cu.bash_red("Test")
        self.assertEqual(result, "\033[91mTest\033[0m")

    def test_bold(self):
        """Test bash_bold output."""
        result = cu.bash_bold("Test")
        self.assertEqual(result, "\033[1mTest\033[0m")

    def test_underline(self):
        """Test bash_underline output."""
        result = cu.bash_underline("Test")
        self.assertEqual(result, "\033[4mTest\033[0m")

#########################################################################################################
# Script entry point


if __name__ == "__main__":  # pragma: no cover
    unittest.main(verbosity=2)



"""
This file contains utilities that are useful for caching
"""

###########################################################################################
# Initial imports:

import os
import pickle
from functools import wraps

###########################################################################################

def cache_input(func, check_input=True):
    """
    Cache a function's positional and keyword arguments on disk.

    The decorator expects ``cache_dir`` and ``root_name`` keyword arguments on
    the wrapped function. It persists the arguments to
    ``<cache_dir>/<root_name>_function_cache.plk`` and reloads them on
    subsequent calls, optionally validating they match the current inputs.

    :param func: function to decorate.
    :param check_input: whether to verify cached inputs match current values.
    :returns: wrapped function that reuses cached arguments when available.
    :raises ValueError: if required cache kwargs are missing or inputs differ
        from the cached values when ``check_input`` is True.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        # Retrieve caching essentials from function's keyword arguments
        cache_dir = kwargs.get('cache_dir', None)
        root_name = kwargs.get('root_name', None)

        # If no cache_dir or root_name is provided, raise an error
        if cache_dir is None or root_name is None:
            raise ValueError('cache_dir and root_name must be provided for input to be cached')

        # Ensure the cache directory exists
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)

        # Define the cache file path
        cache_file = os.path.join(cache_dir, root_name + '_function_cache.plk')

        # If the cache file exists, load the cached arguments
        if os.path.exists(cache_file):
            with open(cache_file, 'rb') as f:
                cached_args, cached_kwargs = pickle.load(f)
            # check whether the cached arguments are the same as the current ones:
            if check_input:
                if args != () and args != cached_args:
                    raise ValueError(
                        f"Cached arguments are different from current ones. "
                        f"args: {args}, cached_args: {cached_args}. "
                        f"Please remove the cache file and run the function again."
                    )
                if kwargs != {}:
                    for key in kwargs:
                        if kwargs[key] != cached_kwargs[key]:
                            raise ValueError(
                                f"Cached arguments are different from current ones. "
                                f"Error in key: {key}. "
                                f"Provided: {kwargs[key]}, Cached: {cached_kwargs[key]}. "
                                f"Please remove the cache file and run the function again."
                            )
            # if all checks pass, return the cached result:
            args = cached_args
            kwargs = cached_kwargs
        else:
            with open(cache_file, 'wb') as f:
                pickle.dump((args, kwargs), f)

        # Call the original function with the arguments
        return func(*args, **kwargs)

    return wrapper



"""Package metadata and convenience constants."""

import warnings

__author__ = 'Marco Raveri'
__version__ = "1.1.0"
__url__ = "https://tensiometer.readthedocs.io"

warnings.filterwarnings(
    "ignore",
    message="In the future `np.object` will be defined as the corresponding NumPy scalar.",
    category=FutureWarning,
)
